from typing import Any

from .interfaces import Dialect
from .. import event

class ConnectionEvents(event.Events):
    def before_execute(
        self,
        conn: Any,
        clauseelement: Any,
        multiparams: Any,
        params: Any,
        execution_options: Any,
    ) -> None: ...
    def after_execute(
        self,
        conn: Any,
        clauseelement: Any,
        multiparams: Any,
        params: Any,
        execution_options: Any,
        result: Any,
    ) -> None: ...
    def before_cursor_execute(
        self,
        conn: Any,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Any,
        executemany: Any,
    ) -> None: ...
    def after_cursor_execute(
        self,
        conn: Any,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Any,
        executemany: Any,
    ) -> None: ...
    def handle_error(self, exception_context: Any) -> None: ...
    def engine_connect(self, conn: Any, branch: Any) -> None: ...
    def set_connection_execution_options(
        self, conn: Any, opts: Any
    ) -> None: ...
    def set_engine_execution_options(self, engine: Any, opts: Any) -> None: ...
    def engine_disposed(self, engine: Any) -> None: ...
    def begin(self, conn: Any) -> None: ...
    def rollback(self, conn: Any) -> None: ...
    def commit(self, conn: Any) -> None: ...
    def savepoint(self, conn: Any, name: Any) -> None: ...
    def rollback_savepoint(
        self, conn: Any, name: Any, context: Any
    ) -> None: ...
    def release_savepoint(
        self, conn: Any, name: Any, context: Any
    ) -> None: ...
    def begin_twophase(self, conn: Any, xid: Any) -> None: ...
    def prepare_twophase(self, conn: Any, xid: Any) -> None: ...
    def rollback_twophase(
        self, conn: Any, xid: Any, is_prepared: Any
    ) -> None: ...
    def commit_twophase(
        self, conn: Any, xid: Any, is_prepared: Any
    ) -> None: ...

class DialectEvents(event.Events):
    def do_connect(
        self, dialect: Dialect, conn_rec: Any, cargs: Any, cparams: Any
    ) -> None: ...
    def do_executemany(
        self, cursor: Any, statement: Any, parameters: Any, context: Any
    ) -> None: ...
    def do_execute_no_params(
        self, cursor: Any, statement: Any, context: Any
    ) -> None: ...
    def do_execute(
        self, cursor: Any, statement: Any, parameters: Any, context: Any
    ) -> None: ...
    def do_setinputsizes(
        self,
        inputsizes: Any,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Any,
    ) -> None: ...



from typing import Any
from typing import Optional

rootlogger: Any

def class_logger(cls): ...

class Identified:
    logging_name: Any = ...

class InstanceLogger:
    echo: Any = ...
    logger: Any = ...
    def __init__(self, echo: Any, name: Any) -> None: ...
    def debug(self, msg: Any, *args: Any, **kwargs: Any) -> None: ...
    def info(self, msg: Any, *args: Any, **kwargs: Any) -> None: ...
    def warning(self, msg: Any, *args: Any, **kwargs: Any) -> None: ...
    warn: Any = ...
    def error(self, msg: Any, *args: Any, **kwargs: Any) -> None: ...
    def exception(self, msg: Any, *args: Any, **kwargs: Any) -> None: ...
    def critical(self, msg: Any, *args: Any, **kwargs: Any) -> None: ...
    def log(self, level: Any, msg: Any, *args: Any, **kwargs: Any) -> None: ...
    def isEnabledFor(self, level: Any): ...
    def getEffectiveLevel(self): ...

def instance_logger(instance: Any, echoflag: Optional[Any] = ...) -> None: ...

class echo_property:
    __doc__: str = ...
    def __get__(self, instance: Any, owner: Any): ...
    def __set__(self, instance: Any, value: Any) -> None: ...



from typing import Any

from .base import BIGINT as BIGINT
from .base import BINARY as BINARY
from .base import BIT as BIT
from .base import CHAR as CHAR
from .base import DATE as DATE
from .base import DATETIME as DATETIME
from .base import DATETIME2 as DATETIME2
from .base import DATETIMEOFFSET as DATETIMEOFFSET
from .base import DECIMAL as DECIMAL
from .base import FLOAT as FLOAT
from .base import IMAGE as IMAGE
from .base import INTEGER as INTEGER
from .base import JSON as JSON
from .base import MONEY as MONEY
from .base import NCHAR as NCHAR
from .base import NTEXT as NTEXT
from .base import NUMERIC as NUMERIC
from .base import NVARCHAR as NVARCHAR
from .base import REAL as REAL
from .base import ROWVERSION as ROWVERSION
from .base import SMALLDATETIME as SMALLDATETIME
from .base import SMALLINT as SMALLINT
from .base import SMALLMONEY as SMALLMONEY
from .base import SQL_VARIANT as SQL_VARIANT
from .base import TEXT as TEXT
from .base import TIME as TIME
from .base import TIMESTAMP as TIMESTAMP
from .base import TINYINT as TINYINT
from .base import try_cast as try_cast
from .base import UNIQUEIDENTIFIER as UNIQUEIDENTIFIER
from .base import VARBINARY as VARBINARY
from .base import VARCHAR as VARCHAR
from .base import XML as XML

dialect: Any



from typing import Any
from typing import ClassVar
from typing import Collection
from typing import Iterable
from typing import NoReturn
from typing import Optional
from typing import Set
from typing import Type
from typing import TypeVar
from typing import Union

from . import dml
from . import roles
from . import sqltypes
from . import type_api
from .annotation import Annotated
from .annotation import SupportsCloneAnnotations
from .base import CacheableOptions
from .base import ColumnCollection
from .base import ColumnSet
from .base import CompileState
from .base import Executable
from .base import Generative
from .base import HasCompileState
from .base import HasMemoized
from .base import Immutable
from .base import ImmutableColumnCollection
from .elements import BindParameter
from .elements import BooleanClauseList
from .elements import ClauseElement
from .elements import ColumnElement
from .elements import GroupedElement
from .elements import Grouping
from .elements import Label
from .elements import TableValuedColumn
from .elements import UnaryExpression
from .schema import Column
from .schema import ForeignKey
from .schema import Table
from .. import util
from ..engine import Connection
from ..engine import Engine
from ..util import langhelpers

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)

_TE = TypeVar("_TE", bound=type_api.TypeEngine[Any])
_S = TypeVar("_S", bound=Selectable)
_HP = TypeVar("_HP", bound=HasPrefixes)
_HS = TypeVar("_HS", bound=HasSuffixes)
_HH = TypeVar("_HH", bound=HasHints)
_VA = TypeVar("_VA", bound=Values)
_SS = TypeVar("_SS", bound=ScalarSelect)
_SSG = TypeVar("_SSG", bound=SelectStatementGrouping)
_GS = TypeVar("_GS", bound=GenerativeSelect)
_SEL = TypeVar("_SEL", bound=Select)
_EX = TypeVar("_EX", bound=Exists)
_TS = TypeVar("_TS", bound=TextualSelect)
_J = TypeVar("_J", bound=Join)

class _OffsetLimitParam(BindParameter[_TE]):
    inherit_cache: bool = ...

def subquery(alias: str, *args: Any, **kwargs: Any) -> Alias: ...

class ReturnsRows(roles.ReturnsRowsRole, ClauseElement):
    @property
    def selectable(self) -> Any: ...
    @property
    def exported_columns(self) -> ColumnCollection[ColumnElement[Any]]: ...

class Selectable(ReturnsRows):
    __visit_name__: str = ...
    is_selectable: bool = ...
    @property
    def selectable(self: _S) -> _S: ...
    def lateral(self, name: Optional[str] = ...) -> Lateral: ...
    def replace_selectable(self: _S, old: Any, alias: Any) -> _S: ...
    def corresponding_column(
        self, column: Column[Any], require_embedded: bool = ...
    ) -> Optional[Column[Any]]: ...

class HasPrefixes:
    def prefix_with(self: _HP, *expr: Any, **kw: Any) -> _HP: ...

class HasSuffixes:
    def suffix_with(self: _HS, *expr: Any, **kw: Any) -> _HS: ...

class HasHints:
    def with_statement_hint(
        self: _HH, text: Any, dialect_name: str = ...
    ) -> _HH: ...
    def with_hint(
        self: _HH, selectable: Any, text: Any, dialect_name: str = ...
    ) -> _HH: ...

class FromClause(roles.AnonymizedFromClauseRole, Selectable):
    __visit_name__: str = ...
    named_with_column: bool = ...
    schema: Any = ...
    is_selectable: bool = ...
    def select(
        self, whereclause: Optional[Any] = ..., **kwargs: Any
    ) -> Select: ...
    def join(
        self,
        right: FromClause,
        onclause: Optional[ClauseElement] = ...,
        isouter: bool = ...,
        full: bool = ...,
    ) -> Join: ...
    def outerjoin(
        self,
        right: FromClause,
        onclause: Optional[ClauseElement] = ...,
        full: bool = ...,
    ) -> Join: ...
    def alias(self, name: Optional[str] = ..., flat: bool = ...) -> Alias: ...
    def table_valued(self) -> TableValuedColumn[sqltypes.TableValueType]: ...
    def tablesample(
        self,
        sampling: Any,
        name: Optional[Any] = ...,
        seed: Optional[Any] = ...,
    ) -> TableSample: ...
    def is_derived_from(self, fromclause: FromClause) -> bool: ...
    @property
    def description(self) -> util.text_type: ...
    @property
    def exported_columns(
        self,
    ) -> ImmutableColumnCollection[ColumnElement[Any]]: ...
    @util.memoized_property
    def columns(self) -> ImmutableColumnCollection[ColumnElement[Any]]: ...
    @property
    def entity_namespace(
        self,
    ) -> ImmutableColumnCollection[ColumnElement[Any]]: ...
    @util.memoized_property
    def primary_key(self) -> ColumnSet[ColumnElement[Any]]: ...
    @util.memoized_property
    def foreign_keys(self) -> Set[ForeignKey]: ...
    @util.memoized_property
    def c(self) -> ImmutableColumnCollection[ColumnElement[Any]]: ...

LABEL_STYLE_NONE: langhelpers._symbol
LABEL_STYLE_TABLENAME_PLUS_COL: langhelpers._symbol
LABEL_STYLE_DISAMBIGUATE_ONLY: langhelpers._symbol
LABEL_STYLE_DEFAULT: langhelpers._symbol

class Join(roles.DMLTableRole, FromClause):
    __visit_name__: str = ...
    left: Any = ...
    right: Any = ...
    onclause: Any = ...
    isouter: bool = ...
    full: bool = ...
    def __init__(
        self,
        left: Union[Type[Any], roles.FromClauseRole],
        right: Union[Type[Any], roles.FromClauseRole],
        onclause: Optional[Any] = ...,
        isouter: bool = ...,
        full: bool = ...,
    ) -> None: ...
    @classmethod
    def _create_outerjoin(
        cls: Type[_J],
        left: Union[Type[Any], roles.FromClauseRole],
        right: Union[Type[Any], roles.FromClauseRole],
        onclause: Optional[Any] = ...,
        full: bool = ...,
    ) -> _J: ...
    @classmethod
    def _create_join(
        cls: Type[_J],
        left: Union[Type[Any], roles.FromClauseRole],
        right: Union[Type[Any], roles.FromClauseRole],
        onclause: Optional[Any] = ...,
        isouter: bool = ...,
        full: bool = ...,
    ) -> _J: ...
    @property
    def description(self) -> util.text_type: ...
    def is_derived_from(self, fromclause: FromClause) -> bool: ...
    def self_group(self, against: Optional[Any] = ...) -> FromGrouping: ...
    @classmethod
    def _join_condition(
        cls,
        a: Union[Type[Any], roles.FromClauseRole],
        b: Union[Type[Any], roles.FromClauseRole],
        a_subset: Optional[Any] = ...,
        consider_as_foreign_keys: Optional[Collection[Table]] = ...,
    ) -> ClauseElement: ...
    def select(
        self, whereclause: Optional[Any] = ..., **kwargs: Any
    ) -> Select: ...
    @property
    def bind(self) -> Optional[Union[Engine, Connection]]: ...
    def alias(  # type: ignore[override]
        self, name: Optional[str] = ..., flat: bool = ...
    ) -> Union[Alias, Join]: ...

class NoInit:
    def __init__(self, *arg: Any, **kw: Any) -> None: ...

class AliasedReturnsRows(NoInit, FromClause):
    named_with_column: bool = ...
    @property
    def description(self) -> util.text_type: ...
    @property
    def original(self) -> ClauseElement: ...
    def is_derived_from(self, fromclause: FromClause) -> bool: ...
    @property
    def bind(self) -> Optional[Union[Engine, Connection]]: ...

class Alias(roles.DMLTableRole, AliasedReturnsRows):
    __visit_name__: str = ...
    inherit_cache: bool = ...
    @classmethod
    def _factory(
        cls,
        selectable: Selectable,
        name: Optional[str] = ...,
        flat: bool = ...,
    ) -> Alias: ...

class TableValuedAlias(Alias):
    __visit_name__: str = ...
    @HasMemoized.memoized_attribute
    def column(self) -> TableValuedColumn[Any]: ...
    def alias(self, name: Optional[Any] = ...) -> TableValuedAlias: ...  # type: ignore[override]
    def lateral(self, name: Optional[Any] = ...) -> TableValuedAlias: ...  # type: ignore[override]
    def render_derived(
        self, name: Optional[Any] = ..., with_types: bool = ...
    ) -> TableValuedAlias: ...

class Lateral(AliasedReturnsRows):
    __visit_name__: str = ...
    inherit_cache: bool = ...
    @classmethod
    def _factory(
        cls, selectable: Selectable, name: Optional[str] = ...
    ) -> Lateral: ...

class TableSample(AliasedReturnsRows):
    __visit_name__: str = ...
    @classmethod
    def _factory(
        cls,
        selectable: Selectable,
        sampling: float,
        name: Optional[str] = ...,
        seed: Optional[Any] = ...,
    ) -> TableSample: ...

class CTE(Generative, HasPrefixes, HasSuffixes, AliasedReturnsRows):
    __visit_name__: str = ...
    @classmethod
    def _factory(
        cls,
        selectable: Any,
        name: Optional[str] = ...,
        recursive: bool = ...,
    ) -> CTE: ...
    def alias(self, name: Optional[str] = ..., flat: bool = ...) -> CTE: ...  # type: ignore[override]
    def union(self, other: Any) -> CTE: ...
    def union_all(self, other: Any) -> CTE: ...

class HasCTE(roles.HasCTERole):
    def cte(self, name: Optional[str] = ..., recursive: bool = ...) -> CTE: ...

class Subquery(AliasedReturnsRows):
    __visit_name__: str = ...
    inherit_cache: bool = ...
    def as_scalar(self) -> ScalarSelect: ...

class FromGrouping(GroupedElement, FromClause):
    element: Any = ...
    def __init__(self, element: Any) -> None: ...
    @property
    def columns(self) -> ImmutableColumnCollection[ColumnElement[Any]]: ...  # type: ignore[override]
    @property
    def primary_key(self) -> ColumnSet[ColumnElement[Any]]: ...  # type: ignore[override]
    @property
    def foreign_keys(self) -> Set[ForeignKey]: ...  # type: ignore[override]
    def is_derived_from(self, element: Any) -> bool: ...
    def alias(self, **kw: Any) -> FromGrouping: ...  # type: ignore[override]

class TableClause(roles.DMLTableRole, Immutable, FromClause):
    __visit_name__: str = ...
    named_with_column: bool = ...
    implicit_returning: bool = ...
    name: Any = ...
    primary_key: Any = ...
    foreign_keys: Any = ...
    schema: Any = ...
    def __init__(
        self, name: str, *columns: Column[Any], **kw: Any
    ) -> None: ...
    @util.memoized_property
    def description(self) -> util.text_type: ...  # type: ignore[override]
    def append_column(self, c: Column[Any], **kw: Any) -> None: ...
    def insert(
        self, values: Optional[Any] = ..., inline: bool = ..., **kwargs: Any
    ) -> dml.Insert: ...
    def update(
        self,
        whereclause: Optional[Any] = ...,
        values: Optional[Any] = ...,
        inline: bool = ...,
        **kwargs: Any,
    ) -> dml.Update: ...
    def delete(
        self, whereclause: Optional[Any] = ..., **kwargs: Any
    ) -> dml.Delete: ...

class ForUpdateArg(ClauseElement):
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __hash__(self) -> int: ...
    nowait: bool = ...
    read: bool = ...
    skip_locked: bool = ...
    key_share: bool = ...
    of: Any = ...
    def __init__(
        self,
        nowait: bool = ...,
        read: bool = ...,
        of: Optional[Any] = ...,
        skip_locked: bool = ...,
        key_share: bool = ...,
    ) -> None: ...

class Values(Generative, FromClause):
    named_with_column: bool = ...
    __visit_name__: str = ...
    name: Any = ...
    literal_binds: Any = ...
    def __init__(self, *columns: Column[Any], **kw: Any) -> None: ...
    def alias(self: _VA, name: Any, **kw: Any) -> _VA: ...  # type: ignore[override]
    def lateral(self: _VA, name: Optional[Any] = ...) -> _VA: ...  # type: ignore[override]
    def data(self: _VA, values: Any) -> _VA: ...

class SelectBase(
    roles.SelectStatementRole,
    roles.DMLSelectRole,
    roles.CompoundElementRole,
    roles.InElementRole,
    HasCTE,
    Executable,
    SupportsCloneAnnotations,
    Selectable,
):
    is_select: bool = ...
    @property
    def selected_columns(
        self,
    ) -> ImmutableColumnCollection[ColumnElement[Any]]: ...
    @property
    def exported_columns(
        self,
    ) -> ImmutableColumnCollection[ColumnElement[Any]]: ...
    @property
    def c(self) -> ColumnCollection[ColumnElement[Any]]: ...
    @property
    def columns(self) -> ColumnCollection[ColumnElement[Any]]: ...
    def select(self, *arg: Any, **kw: Any) -> Select: ...
    def as_scalar(self) -> ScalarSelect: ...
    def exists(self) -> Exists: ...
    def scalar_subquery(self) -> ScalarSelect: ...
    def label(self, name: Any) -> Label[Any]: ...
    def lateral(self, name: Optional[Any] = ...) -> Lateral: ...
    def subquery(self, name: Optional[Any] = ...) -> Subquery: ...
    def alias(
        self, name: Optional[Any] = ..., flat: bool = ...
    ) -> Subquery: ...

class SelectStatementGrouping(GroupedElement, SelectBase):
    __visit_name__: str = ...
    element: Any = ...
    def __init__(self, element: SelectBase) -> None: ...
    def get_label_style(self) -> langhelpers._symbol: ...
    def set_label_style(
        self, label_style: langhelpers._symbol
    ) -> SelectStatementGrouping: ...
    @property
    def select_statement(self) -> SelectBase: ...
    def self_group(self: _SSG, against: Optional[Any] = ...) -> _SSG: ...
    @property
    def selected_columns(
        self,
    ) -> ImmutableColumnCollection[ColumnElement[Any]]: ...

class DeprecatedSelectBaseGenerations:
    def append_order_by(self, *clauses: Any) -> None: ...
    def append_group_by(self, *clauses: Any) -> None: ...

class GenerativeSelect(DeprecatedSelectBaseGenerations, SelectBase):
    def __init__(
        self,
        _label_style: Any = ...,
        use_labels: bool = ...,
        limit: Optional[Any] = ...,
        offset: Optional[Any] = ...,
        order_by: Optional[Any] = ...,
        group_by: Optional[Any] = ...,
        bind: Optional[Any] = ...,
    ) -> None: ...
    def with_for_update(
        self: _GS,
        nowait: bool = ...,
        read: bool = ...,
        of: Optional[Any] = ...,
        skip_locked: bool = ...,
        key_share: bool = ...,
    ) -> _GS: ...
    def get_label_style(self) -> langhelpers._symbol: ...
    def set_label_style(self: _GS, style: langhelpers._symbol) -> _GS: ...
    def apply_labels(self: _GS) -> _GS: ...
    def limit(self: _GS, limit: Any) -> _GS: ...
    def fetch(
        self: _GS, count: Any, with_ties: bool = ..., percent: bool = ...
    ) -> _GS: ...
    def offset(self: _GS, offset: Any) -> _GS: ...
    def slice(self: _GS, start: Any, stop: Any) -> _GS: ...
    def order_by(self: _GS, *clauses: Any) -> _GS: ...
    def group_by(self: _GS, *clauses: Any) -> _GS: ...

class CompoundSelectState(CompileState): ...

class CompoundSelect(HasCompileState, GenerativeSelect):
    __visit_name__: str = ...
    UNION: ClassVar[langhelpers._symbol] = ...
    UNION_ALL: ClassVar[langhelpers._symbol] = ...
    EXCEPT: ClassVar[langhelpers._symbol] = ...
    EXCEPT_ALL: ClassVar[langhelpers._symbol] = ...
    INTERSECT: ClassVar[langhelpers._symbol] = ...
    INTERSECT_ALL: ClassVar[langhelpers._symbol] = ...
    keyword: langhelpers._symbol = ...
    selects: Any = ...
    bind: Optional[Union[Engine, Connection]] = ...
    def __init__(
        self, keyword: langhelpers._symbol, *selects: Any, **kwargs: Any
    ) -> None: ...
    @classmethod
    def _create_union(cls, *selects: Any, **kwargs: Any) -> CompoundSelect: ...
    @classmethod
    def _create_union_all(
        cls, *selects: Any, **kwargs: Any
    ) -> CompoundSelect: ...
    @classmethod
    def _create_except(
        cls, *selects: Any, **kwargs: Any
    ) -> CompoundSelect: ...
    @classmethod
    def _create_except_all(
        cls, *selects: Any, **kwargs: Any
    ) -> CompoundSelect: ...
    @classmethod
    def _create_intersect(
        cls, *selects: Any, **kwargs: Any
    ) -> CompoundSelect: ...
    @classmethod
    def _create_intersect_all(
        cls, *selects: Any, **kwargs: Any
    ) -> CompoundSelect: ...
    def self_group(
        self, against: Optional[Any] = ...
    ) -> SelectStatementGrouping: ...
    def is_derived_from(self, fromclause: Any) -> bool: ...
    @property
    def selected_columns(
        self,
    ) -> ImmutableColumnCollection[ColumnElement[Any]]: ...

class DeprecatedSelectGenerations:
    def append_correlation(self, fromclause: Any) -> None: ...
    def append_column(self, column: Any) -> None: ...
    def append_prefix(self, clause: Any) -> None: ...
    def append_whereclause(self, whereclause: Any) -> None: ...
    def append_having(self, having: Any) -> None: ...
    def append_from(self, fromclause: Any) -> None: ...

class SelectState(util.MemoizedSlots, CompileState):
    class default_select_compile_options(CacheableOptions): ...
    statement: Any = ...
    from_clauses: Any = ...
    froms: Any = ...
    columns_plus_names: Any = ...
    def __init__(self, statement: Any, compiler: Any, **kw: Any) -> None: ...
    @classmethod
    def get_column_descriptions(cls, statement: Any) -> NoReturn: ...
    @classmethod
    def from_statement(
        cls, statement: Any, from_statement: Any
    ) -> NoReturn: ...
    @classmethod
    def determine_last_joined_entity(cls, stmt: Any) -> Optional[Any]: ...
    @classmethod
    def exported_columns_iterator(
        cls, statement: Any
    ) -> Iterable[ColumnElement[Any]]: ...

class _SelectFromElements: ...

class Select(
    HasPrefixes,
    HasSuffixes,
    HasHints,
    HasCompileState,
    DeprecatedSelectGenerations,
    _SelectFromElements,
    GenerativeSelect,
):
    __visit_name__: str = ...
    bind: Optional[Union[Engine, Connection]] = ...
    @classmethod
    def create_legacy_select(
        cls: Type[_SEL],
        columns: Optional[Any] = ...,
        whereclause: Optional[Any] = ...,
        from_obj: Optional[Any] = ...,
        distinct: bool = ...,
        having: Optional[Any] = ...,
        correlate: bool = ...,
        prefixes: Optional[Any] = ...,
        suffixes: Optional[Any] = ...,
        **kwargs: Any,
    ) -> _SEL: ...
    @classmethod
    def _create(cls, *args: Any, **kw: Any) -> Select: ...
    def __init__(self) -> None: ...
    def filter(self: _SEL, *criteria: Any) -> _SEL: ...
    def filter_by(self: _SEL, **kwargs: Any) -> _SEL: ...
    @property
    def column_descriptions(self) -> Any: ...
    def from_statement(self, statement: Any) -> Any: ...
    def join(
        self: _SEL,
        target: Any,
        onclause: Optional[Any] = ...,
        isouter: bool = ...,
        full: bool = ...,
    ) -> _SEL: ...
    def outerjoin_from(
        self: _SEL,
        from_: Any,
        target: Any,
        onclause: Optional[Any] = ...,
        full: bool = ...,
    ) -> _SEL: ...
    def join_from(
        self: _SEL,
        from_: Any,
        target: Any,
        onclause: Optional[Any] = ...,
        isouter: bool = ...,
        full: bool = ...,
    ) -> _SEL: ...
    def outerjoin(
        self: _SEL,
        target: Any,
        onclause: Optional[Any] = ...,
        full: bool = ...,
    ) -> _SEL: ...
    @property
    def froms(self) -> Any: ...
    @property
    def inner_columns(self) -> Any: ...
    def is_derived_from(self, fromclause: Any) -> bool: ...
    def get_children(self, **kwargs: Any) -> Iterable[Any]: ...  # type: ignore[override]
    def add_columns(self: _SEL, *columns: Any) -> _SEL: ...
    def column(self: _SEL, column: Any) -> _SEL: ...
    def reduce_columns(self: _SEL, only_synonyms: bool = ...) -> _SEL: ...
    def with_only_columns(self: _SEL, *columns: Any) -> _SEL: ...
    @property
    def whereclause(self) -> BooleanClauseList[Any]: ...
    def where(self: _SEL, *whereclause: Any) -> _SEL: ...
    def having(self: _SEL, having: Any) -> _SEL: ...
    def distinct(self: _SEL, *expr: Any) -> _SEL: ...
    def select_from(self: _SEL, *froms: Any) -> _SEL: ...
    def correlate(self: _SEL, *fromclauses: Any) -> _SEL: ...
    def correlate_except(self: _SEL, *fromclauses: Any) -> _SEL: ...
    @HasMemoized.memoized_attribute
    def selected_columns(self) -> ImmutableColumnCollection[ColumnElement[Any]]: ...  # type: ignore[override]
    def self_group(
        self: _SEL, against: Optional[Any] = ...
    ) -> Union[_SEL, SelectStatementGrouping]: ...
    def union(self, other: Any, **kwargs: Any) -> CompoundSelect: ...
    def union_all(self, other: Any, **kwargs: Any) -> CompoundSelect: ...
    def except_(self, other: Any, **kwargs: Any) -> CompoundSelect: ...
    def except_all(self, other: Any, **kwargs: Any) -> CompoundSelect: ...
    def intersect(self, other: Any, **kwargs: Any) -> CompoundSelect: ...
    def intersect_all(self, other: Any, **kwargs: Any) -> CompoundSelect: ...

class ScalarSelect(roles.InElementRole, Generative, Grouping[Any]):
    inherit_cache: bool = ...
    element: Any = ...
    type: type_api.TypeEngine[Any] = ...
    def __init__(self, element: Any) -> None: ...
    @property
    def columns(self) -> NoReturn: ...
    @property
    def c(self) -> NoReturn: ...
    def where(self: _SS, crit: Any) -> _SS: ...
    def self_group(self: _SS, **kwargs: Any) -> _SS: ...  # type: ignore[override]
    def correlate(self: _SS, *fromclauses: Any) -> _SS: ...
    def correlate_except(self: _SS, *fromclauses: Any) -> _SS: ...

class Exists(UnaryExpression[sqltypes.Boolean]):
    inherit_cache: bool = ...
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def select(
        self, whereclause: Optional[Any] = ..., **kwargs: Any
    ) -> Select: ...
    def correlate(self: _EX, *fromclause: Any) -> _EX: ...
    def correlate_except(self: _EX, *fromclause: Any) -> _EX: ...
    def select_from(self: _EX, *froms: Any) -> _EX: ...
    def where(self: _EX, clause: Any) -> _EX: ...

class TextualSelect(SelectBase):
    __visit_name__: str = ...
    is_text: bool = ...
    is_select: bool = ...
    element: Any = ...
    column_args: Any = ...
    positional: Any = ...
    def __init__(
        self, text: Any, columns: Any, positional: bool = ...
    ) -> None: ...
    @HasMemoized.memoized_attribute
    def selected_columns(self) -> ImmutableColumnCollection[ColumnElement[Any]]: ...  # type: ignore[override]
    def bindparams(self: _TS, *binds: Any, **bind_as_values: Any) -> _TS: ...

TextAsFrom = TextualSelect

class AnnotatedFromClause(Annotated):
    def __init__(self, element: Any, values: Any) -> None: ...



from typing import Any
from typing import Optional

from .mysqldb import MySQLDialect_mysqldb as MySQLDialect_mysqldb
from ...util import langhelpers as langhelpers
from ...util import py3k as py3k

class MySQLDialect_pymysql(MySQLDialect_mysqldb):
    driver: str = ...
    description_encoding: Any = ...
    supports_unicode_statements: bool = ...
    supports_unicode_binds: bool = ...
    def supports_server_side_cursors(self): ...
    @classmethod
    def dbapi(cls): ...
    def create_connect_args(
        self, url: Any, _translate_args: Optional[Any] = ...
    ): ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...

dialect = MySQLDialect_pymysql



from typing import Any
from typing import Optional

from . import characteristics as characteristics
from . import interfaces as interfaces
from .. import event as event
from .. import exc as exc
from .. import pool as pool
from .. import processors as processors
from .. import types as sqltypes
from .. import util as util
from ..sql import compiler as compiler
from ..sql import expression as expression
from ..sql.elements import quoted_name as quoted_name

AUTOCOMMIT_REGEXP: Any
SERVER_SIDE_CURSOR_RE: Any
CACHE_HIT: Any
CACHE_MISS: Any
CACHING_DISABLED: Any
NO_CACHE_KEY: Any

class DefaultDialect(interfaces.Dialect):
    statement_compiler: Any = ...
    ddl_compiler: Any = ...
    type_compiler: Any = ...
    preparer: Any = ...
    supports_alter: bool = ...
    supports_comments: bool = ...
    inline_comments: bool = ...
    use_setinputsizes: bool = ...
    default_sequence_base: int = ...
    execute_sequence_format: Any = ...
    supports_views: bool = ...
    supports_sequences: bool = ...
    sequences_optional: bool = ...
    preexecute_autoincrement_sequences: bool = ...
    postfetch_lastrowid: bool = ...
    implicit_returning: bool = ...
    full_returning: bool = ...
    insert_executemany_returning: bool = ...
    cte_follows_insert: bool = ...
    supports_native_enum: bool = ...
    supports_native_boolean: bool = ...
    non_native_boolean_check_constraint: bool = ...
    supports_simple_order_by_label: bool = ...
    tuple_in_values: bool = ...
    connection_characteristics: Any = ...
    engine_config_types: Any = ...
    supports_native_decimal: bool = ...
    supports_unicode_statements: bool = ...
    supports_unicode_binds: bool = ...
    returns_unicode_strings: Any = ...
    description_encoding: Any = ...
    name: str = ...
    max_identifier_length: int = ...
    isolation_level: Any = ...
    max_index_name_length: Any = ...
    max_constraint_name_length: Any = ...
    supports_sane_rowcount: bool = ...
    supports_sane_multi_rowcount: bool = ...
    colspecs: Any = ...
    default_paramstyle: str = ...
    supports_default_values: bool = ...
    supports_empty_insert: bool = ...
    supports_multivalues_insert: bool = ...
    supports_is_distinct_from: bool = ...
    supports_server_side_cursors: bool = ...
    server_side_cursors: bool = ...
    supports_for_update_of: bool = ...
    server_version_info: Any = ...
    default_schema_name: Any = ...
    construct_arguments: Any = ...
    requires_name_normalize: bool = ...
    reflection_options: Any = ...
    dbapi_exception_translation_map: Any = ...
    is_async: bool = ...
    CACHE_HIT: Any = ...
    CACHE_MISS: Any = ...
    CACHING_DISABLED: Any = ...
    NO_CACHE_KEY: Any = ...
    convert_unicode: Any = ...
    encoding: Any = ...
    positional: bool = ...
    dbapi: Any = ...
    paramstyle: Any = ...
    identifier_preparer: Any = ...
    case_sensitive: Any = ...
    label_length: Any = ...
    compiler_linting: Any = ...
    def __init__(
        self,
        convert_unicode: bool = ...,
        encoding: str = ...,
        paramstyle: Optional[Any] = ...,
        dbapi: Optional[Any] = ...,
        implicit_returning: Optional[Any] = ...,
        case_sensitive: bool = ...,
        supports_native_boolean: Optional[Any] = ...,
        max_identifier_length: Optional[Any] = ...,
        label_length: Optional[Any] = ...,
        compiler_linting: Any = ...,
        server_side_cursors: bool = ...,
        **kwargs: Any,
    ) -> None: ...
    @property
    def dialect_description(self): ...
    @property
    def supports_sane_rowcount_returning(self): ...
    @classmethod
    def get_pool_class(cls, url: Any): ...
    def get_dialect_pool_class(self, url: Any): ...
    @classmethod
    def load_provisioning(cls) -> None: ...
    default_isolation_level: Any = ...
    def initialize(self, connection: Any) -> None: ...
    def on_connect(self) -> None: ...
    def get_default_isolation_level(self, dbapi_conn: Any): ...
    def type_descriptor(self, typeobj: Any): ...
    def has_index(
        self,
        connection: Any,
        table_name: Any,
        index_name: Any,
        schema: Optional[Any] = ...,
    ): ...
    def validate_identifier(self, ident: Any) -> None: ...
    def connect(self, *cargs: Any, **cparams: Any): ...
    def create_connect_args(self, url: Any): ...
    def set_engine_execution_options(self, engine: Any, opts: Any) -> None: ...
    def set_connection_execution_options(
        self, connection: Any, opts: Any
    ) -> None: ...
    def do_begin(self, dbapi_connection: Any) -> None: ...
    def do_rollback(self, dbapi_connection: Any) -> None: ...
    def do_commit(self, dbapi_connection: Any) -> None: ...
    def do_close(self, dbapi_connection: Any) -> None: ...
    def do_ping(self, dbapi_connection: Any): ...
    def create_xid(self): ...
    def do_savepoint(self, connection: Any, name: Any) -> None: ...
    def do_rollback_to_savepoint(self, connection: Any, name: Any) -> None: ...
    def do_release_savepoint(self, connection: Any, name: Any) -> None: ...
    def do_executemany(
        self,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Optional[Any] = ...,
    ) -> None: ...
    def do_execute(
        self,
        cursor: Any,
        statement: Any,
        parameters: Any,
        context: Optional[Any] = ...,
    ) -> None: ...
    def do_execute_no_params(
        self, cursor: Any, statement: Any, context: Optional[Any] = ...
    ) -> None: ...
    def is_disconnect(self, e: Any, connection: Any, cursor: Any): ...
    def reset_isolation_level(self, dbapi_conn: Any) -> None: ...
    def normalize_name(self, name: Any): ...
    def denormalize_name(self, name: Any): ...

class _RendersLiteral:
    def literal_processor(self, dialect: Any): ...

class _StrDateTime(_RendersLiteral, sqltypes.DateTime): ...
class _StrDate(_RendersLiteral, sqltypes.Date): ...
class _StrTime(_RendersLiteral, sqltypes.Time): ...

class StrCompileDialect(DefaultDialect):
    statement_compiler: Any = ...
    ddl_compiler: Any = ...
    type_compiler: Any = ...
    preparer: Any = ...
    supports_sequences: bool = ...
    sequences_optional: bool = ...
    preexecute_autoincrement_sequences: bool = ...
    implicit_returning: bool = ...
    supports_native_boolean: bool = ...
    supports_simple_order_by_label: bool = ...
    colspecs: Any = ...

class DefaultExecutionContext(interfaces.ExecutionContext):
    isinsert: bool = ...
    isupdate: bool = ...
    isdelete: bool = ...
    is_crud: bool = ...
    is_text: bool = ...
    isddl: bool = ...
    executemany: bool = ...
    compiled: Any = ...
    statement: Any = ...
    result_column_struct: Any = ...
    returned_default_rows: Any = ...
    execution_options: Any = ...
    include_set_input_sizes: Any = ...
    exclude_set_input_sizes: Any = ...
    cursor_fetch_strategy: Any = ...
    cache_stats: Any = ...
    invoked_statement: Any = ...
    cache_hit: Any = ...
    def identifier_preparer(self): ...
    def engine(self): ...
    def postfetch_cols(self): ...
    def prefetch_cols(self): ...
    def returning_cols(self) -> None: ...
    def no_parameters(self): ...
    def should_autocommit(self): ...
    @property
    def connection(self): ...
    def should_autocommit_text(self, statement: Any): ...
    def create_cursor(self): ...
    def create_default_cursor(self): ...
    def create_server_side_cursor(self) -> None: ...
    def pre_exec(self) -> None: ...
    def get_out_parameter_values(self, names: Any) -> None: ...
    def post_exec(self) -> None: ...
    def get_result_processor(self, type_: Any, colname: Any, coltype: Any): ...
    def get_lastrowid(self): ...
    def handle_dbapi_exception(self, e: Any) -> None: ...
    @property
    def rowcount(self): ...
    def supports_sane_rowcount(self): ...
    def supports_sane_multi_rowcount(self): ...
    def inserted_primary_key_rows(self): ...
    def lastrow_has_defaults(self): ...
    current_parameters: Any = ...
    def get_current_parameters(
        self, isolate_multiinsert_groups: bool = ...
    ): ...
    def get_insert_default(self, column: Any): ...
    def get_update_default(self, column: Any): ...



from types import ModuleType
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import Optional
from typing import Tuple

from . import roles

elements: ModuleType
lambdas: ModuleType
schema: ModuleType
selectable: ModuleType
sqltypes: ModuleType
traversals: ModuleType

def expect(
    role: Any,
    element: Any,
    apply_propagate_attrs: Optional[Any] = ...,
    argname: Optional[Any] = ...,
    **kw: Any,
) -> Any: ...
def expect_as_key(role: Any, element: Any, **kw: Any) -> Any: ...
def expect_col_expression_collection(
    role: Any, expressions: Iterable[Any]
) -> Iterator[Tuple[Any, Any, str, Any]]: ...

class RoleImpl:
    name: str = ...
    def __init__(self, role_class: Any) -> None: ...

class _Deannotate: ...
class _StringOnly: ...
class _ReturnsStringKey: ...
class _ColumnCoercions: ...
class _NoTextCoercion: ...
class _CoerceLiterals: ...
class LiteralValueImpl(RoleImpl): ...
class _SelectIsNotFrom: ...
class HasCacheKeyImpl(RoleImpl): ...
class ExpressionElementImpl(_ColumnCoercions, RoleImpl): ...
class BinaryElementImpl(ExpressionElementImpl, RoleImpl): ...
class InElementImpl(RoleImpl): ...
class OnClauseImpl(_CoerceLiterals, _ColumnCoercions, RoleImpl): ...
class WhereHavingImpl(_CoerceLiterals, _ColumnCoercions, RoleImpl): ...
class StatementOptionImpl(_CoerceLiterals, RoleImpl): ...
class ColumnArgumentImpl(_NoTextCoercion, RoleImpl): ...
class ColumnArgumentOrKeyImpl(_ReturnsStringKey, RoleImpl): ...
class StrAsPlainColumnImpl(_CoerceLiterals, RoleImpl): ...
class ByOfImpl(
    _CoerceLiterals, _ColumnCoercions, RoleImpl, roles.ByOfRole
): ...
class OrderByImpl(ByOfImpl, RoleImpl): ...
class GroupByImpl(ByOfImpl, RoleImpl): ...
class DMLColumnImpl(_ReturnsStringKey, RoleImpl): ...
class ConstExprImpl(RoleImpl): ...
class TruncatedLabelImpl(_StringOnly, RoleImpl): ...
class DDLExpressionImpl(_Deannotate, _CoerceLiterals, RoleImpl): ...
class DDLConstraintColumnImpl(_Deannotate, _ReturnsStringKey, RoleImpl): ...
class DDLReferredColumnImpl(DDLConstraintColumnImpl): ...
class LimitOffsetImpl(RoleImpl): ...
class LabeledColumnExprImpl(ExpressionElementImpl): ...
class ColumnsClauseImpl(_SelectIsNotFrom, _CoerceLiterals, RoleImpl): ...
class ReturnsRowsImpl(RoleImpl): ...
class StatementImpl(_NoTextCoercion, RoleImpl): ...
class CoerceTextStatementImpl(_CoerceLiterals, RoleImpl): ...
class SelectStatementImpl(_NoTextCoercion, RoleImpl): ...
class HasCTEImpl(ReturnsRowsImpl): ...
class JoinTargetImpl(RoleImpl): ...
class FromClauseImpl(_SelectIsNotFrom, _NoTextCoercion, RoleImpl): ...
class StrictFromClauseImpl(FromClauseImpl): ...
class AnonymizedFromClauseImpl(StrictFromClauseImpl): ...
class DMLTableImpl(_SelectIsNotFrom, _NoTextCoercion, RoleImpl): ...
class DMLSelectImpl(_NoTextCoercion, RoleImpl): ...
class CompoundElementImpl(_NoTextCoercion, RoleImpl): ...



# fmt: off
from typing import Any

from sqlalchemy import processors as processors
from sqlalchemy import types as sqltypes
from sqlalchemy.connectors.pyodbc import PyODBCConnector as PyODBCConnector
from sqlalchemy.dialects.sybase.base import SybaseDialect as SybaseDialect
from sqlalchemy.dialects.sybase.base import SybaseExecutionContext as SybaseExecutionContext
# fmt: on

class _SybNumeric_pyodbc(sqltypes.Numeric):
    def bind_processor(self, dialect: Any): ...

class SybaseExecutionContext_pyodbc(SybaseExecutionContext):
    def set_ddl_autocommit(self, connection: Any, value: Any) -> None: ...

class SybaseDialect_pyodbc(PyODBCConnector, SybaseDialect):
    execution_ctx_cls: Any = ...
    colspecs: Any = ...
    @classmethod
    def dbapi(cls): ...

dialect = SybaseDialect_pyodbc



from typing import Any
from typing import Optional

from .. import exc as sa_exc
from .. import util as util
from ..exc import MultipleResultsFound as MultipleResultsFound
from ..exc import NoResultFound as NoResultFound

NO_STATE: Any

class StaleDataError(sa_exc.SQLAlchemyError): ...

ConcurrentModificationError = StaleDataError

class FlushError(sa_exc.SQLAlchemyError): ...
class UnmappedError(sa_exc.InvalidRequestError): ...
class ObjectDereferencedError(sa_exc.SQLAlchemyError): ...

class DetachedInstanceError(sa_exc.SQLAlchemyError):
    code: str = ...

class UnmappedInstanceError(UnmappedError):
    def __init__(self, obj: Any, msg: Optional[Any] = ...) -> None: ...
    def __reduce__(self): ...

class UnmappedClassError(UnmappedError):
    def __init__(self, cls: Any, msg: Optional[Any] = ...) -> None: ...
    def __reduce__(self): ...

class ObjectDeletedError(sa_exc.InvalidRequestError):
    def __init__(self, state: Any, msg: Optional[Any] = ...) -> None: ...
    def __reduce__(self): ...

class UnmappedColumnError(sa_exc.InvalidRequestError): ...

class LoaderStrategyException(sa_exc.InvalidRequestError):
    def __init__(
        self,
        applied_to_property_type: Any,
        requesting_property: Any,
        applies_to: Any,
        actual_strategy_type: Any,
        strategy_key: Any,
    ) -> None: ...



from typing import Any
from typing import Optional

class scoped_session:
    session_factory: Any = ...
    registry: Any = ...
    def __init__(
        self, session_factory: Any, scopefunc: Optional[Any] = ...
    ) -> None: ...
    def __call__(self, **kw: Any): ...
    def remove(self) -> None: ...
    def configure(self, **kwargs: Any) -> None: ...
    def query_property(self, query_cls: Optional[Any] = ...): ...

ScopedSession = scoped_session



from typing import Any
from typing import Optional

from ... import types as sqltypes

class JSONPathType(sqltypes.JSON.JSONPathType):
    def bind_processor(self, dialect: Any): ...
    def literal_processor(self, dialect: Any): ...

class JSON(sqltypes.JSON):
    astext_type: Any = ...
    def __init__(
        self, none_as_null: bool = ..., astext_type: Optional[Any] = ...
    ) -> None: ...
    class Comparator(sqltypes.JSON.Comparator):
        @property
        def astext(self): ...
    comparator_factory: Any = ...

class JSONB(JSON):
    __visit_name__: str = ...
    class Comparator(JSON.Comparator):
        def has_key(self, other: Any): ...
        def has_all(self, other: Any): ...
        def has_any(self, other: Any): ...
        def contains(self, other: Any, **kwargs: Any): ...
        def contained_by(self, other: Any): ...
    comparator_factory: Any = ...



# fmt: off
from typing import Any
from typing import Optional

from . import attributes as attributes
from .base import attribute_str as attribute_str
from .base import class_mapper as class_mapper
from .base import InspectionAttr as InspectionAttr
from .base import instance_str as instance_str
from .base import object_mapper as object_mapper
from .base import object_state as object_state
from .base import state_attribute_str as state_attribute_str
from .base import state_class_str as state_class_str
from .base import state_str as state_str
from .interfaces import CriteriaOption as CriteriaOption
from .interfaces import MapperProperty as MapperProperty
from .interfaces import ORMColumnsClauseRole as ORMColumnsClauseRole
from .interfaces import ORMEntityColumnsClauseRole as ORMEntityColumnsClauseRole
from .interfaces import ORMFromClauseRole as ORMFromClauseRole
from .interfaces import PropComparator as PropComparator
from .path_registry import PathRegistry as PathRegistry
from .. import event as event
from .. import inspection as inspection
from .. import sql as sql
from .. import util as util
from ..engine.result import result_tuple as result_tuple
from ..sql import base as sql_base
from ..sql import coercions as coercions
from ..sql import expression as expression
from ..sql import lambdas as lambdas
from ..sql import roles as roles
from ..sql import util as sql_util
from ..sql import visitors as visitors
from ..sql.annotation import SupportsCloneAnnotations as SupportsCloneAnnotations
from ..sql.base import ColumnCollection as ColumnCollection
# fmt: on

all_cascades: Any

class CascadeOptions(frozenset):
    save_update: Any = ...
    delete: Any = ...
    refresh_expire: Any = ...
    merge: Any = ...
    expunge: Any = ...
    delete_orphan: Any = ...
    def __new__(cls, value_list: Any): ...
    @classmethod
    def from_string(cls, arg: Any): ...

def polymorphic_union(
    table_map: Any,
    typecolname: Any,
    aliasname: str = ...,
    cast_nulls: bool = ...,
): ...
def identity_key(*args: Any, **kwargs: Any): ...

class ORMAdapter(sql_util.ColumnAdapter):
    mapper: Any = ...
    aliased_class: Any = ...
    def __init__(
        self,
        entity: Any,
        equivalents: Optional[Any] = ...,
        adapt_required: bool = ...,
        allow_label_resolve: bool = ...,
        anonymize_labels: bool = ...,
    ) -> None: ...

class AliasedClass:
    __name__: Any = ...
    def __init__(
        self,
        mapped_class_or_ac: Any,
        alias: Optional[Any] = ...,
        name: Optional[Any] = ...,
        flat: bool = ...,
        adapt_on_names: bool = ...,
        with_polymorphic_mappers: Any = ...,
        with_polymorphic_discriminator: Optional[Any] = ...,
        base_alias: Optional[Any] = ...,
        use_mapper_path: bool = ...,
        represents_outer_join: bool = ...,
    ) -> None: ...
    def __getattr__(self, key: Any): ...

class AliasedInsp(
    ORMEntityColumnsClauseRole,
    ORMFromClauseRole,
    sql_base.MemoizedHasCacheKey,
    InspectionAttr,
):
    mapper: Any = ...
    selectable: Any = ...
    name: Any = ...
    polymorphic_on: Any = ...
    represents_outer_join: Any = ...
    with_polymorphic_mappers: Any = ...
    def __init__(
        self,
        entity: Any,
        inspected: Any,
        selectable: Any,
        name: Any,
        with_polymorphic_mappers: Any,
        polymorphic_on: Any,
        _base_alias: Any,
        _use_mapper_path: Any,
        adapt_on_names: Any,
        represents_outer_join: Any,
    ) -> None: ...
    @property
    def entity(self): ...
    is_aliased_class: bool = ...
    def __clause_element__(self): ...
    @property
    def entity_namespace(self): ...
    @property
    def class_(self): ...

class LoaderCriteriaOption(CriteriaOption):
    root_entity: Any = ...
    entity: Any = ...
    deferred_where_criteria: bool = ...
    where_criteria: Any = ...
    include_aliases: Any = ...
    propagate_to_loaders: Any = ...
    def __init__(
        self,
        entity_or_base: Any,
        where_criteria: Any,
        loader_only: bool = ...,
        include_aliases: bool = ...,
        propagate_to_loaders: bool = ...,
        track_closure_variables: bool = ...,
    ) -> None: ...
    def process_compile_state(self, compile_state: Any) -> None: ...
    def get_global_criteria(self, attributes: Any) -> None: ...

def aliased(
    element: Any,
    alias: Optional[Any] = ...,
    name: Optional[Any] = ...,
    flat: bool = ...,
    adapt_on_names: bool = ...,
): ...
def with_polymorphic(
    base: Any,
    classes: Any,
    selectable: bool = ...,
    flat: bool = ...,
    polymorphic_on: Optional[Any] = ...,
    aliased: bool = ...,
    innerjoin: bool = ...,
    _use_mapper_path: bool = ...,
    _existing_alias: Optional[Any] = ...,
): ...

class Bundle(ORMColumnsClauseRole, SupportsCloneAnnotations, InspectionAttr):
    single_entity: bool = ...
    is_clause_element: bool = ...
    is_mapper: bool = ...
    is_aliased_class: bool = ...
    is_bundle: bool = ...
    name: Any = ...
    exprs: Any = ...
    c: Any = ...
    def __init__(self, name: Any, *exprs: Any, **kw: Any) -> None: ...
    @property
    def mapper(self): ...
    @property
    def entity(self): ...
    @property
    def entity_namespace(self): ...
    columns: Any = ...
    def __clause_element__(self): ...
    @property
    def clauses(self): ...
    def label(self, name: Any): ...
    def create_row_processor(self, query: Any, procs: Any, labels: Any): ...

class _ORMJoin(expression.Join):
    __visit_name__: Any = ...
    inherit_cache: bool = ...
    onclause: Any = ...
    def __init__(
        self,
        left: Any,
        right: Any,
        onclause: Optional[Any] = ...,
        isouter: bool = ...,
        full: bool = ...,
        _left_memo: Optional[Any] = ...,
        _right_memo: Optional[Any] = ...,
        _extra_criteria: Any = ...,
    ) -> None: ...
    def join(
        self,
        right: Any,
        onclause: Optional[Any] = ...,
        isouter: bool = ...,
        full: bool = ...,
        join_to_left: Optional[Any] = ...,
    ): ...
    def outerjoin(
        self,
        right: Any,
        onclause: Optional[Any] = ...,
        full: bool = ...,
        join_to_left: Optional[Any] = ...,
    ): ...

def join(
    left: Any,
    right: Any,
    onclause: Optional[Any] = ...,
    isouter: bool = ...,
    full: bool = ...,
    join_to_left: Optional[Any] = ...,
): ...
def outerjoin(
    left: Any,
    right: Any,
    onclause: Optional[Any] = ...,
    full: bool = ...,
    join_to_left: Optional[Any] = ...,
): ...
def with_parent(
    instance: Any, prop: Any, from_entity: Optional[Any] = ...
): ...
def has_identity(object_: Any): ...
def was_deleted(object_: Any): ...
def randomize_unitofwork() -> None: ...



# fmt: off
from typing import Any

from sqlalchemy.connectors.mxodbc import MxODBCConnector as MxODBCConnector
from sqlalchemy.dialects.sybase.base import SybaseDialect as SybaseDialect
from sqlalchemy.dialects.sybase.base import SybaseExecutionContext as SybaseExecutionContext
# fmt: on

class SybaseExecutionContext_mxodbc(SybaseExecutionContext): ...

class SybaseDialect_mxodbc(MxODBCConnector, SybaseDialect):
    execution_ctx_cls: Any = ...

dialect = SybaseDialect_mxodbc



from .extensions import AbstractConcreteBase as AbstractConcreteBase
from .extensions import ConcreteBase as ConcreteBase
from .extensions import DeferredReflection as DeferredReflection
from .extensions import instrument_declarative as instrument_declarative
from ...orm.decl_api import as_declarative as as_declarative
from ...orm.decl_api import declarative_base as declarative_base
from ...orm.decl_api import DeclarativeMeta as DeclarativeMeta
from ...orm.decl_api import declared_attr as declared_attr
from ...orm.decl_api import has_inherited_table as has_inherited_table
from ...orm.decl_api import synonym_for as synonym_for



from typing import Any
from typing import Optional

from .. import util as util
from ..orm import backref as backref
from ..orm import interfaces as interfaces
from ..orm import relationship as relationship
from ..schema import ForeignKeyConstraint as ForeignKeyConstraint
from ..sql import and_ as and_

def classname_for_table(base: Any, tablename: Any, table: Any): ...
def name_for_scalar_relationship(
    base: Any, local_cls: Any, referred_cls: Any, constraint: Any
): ...
def name_for_collection_relationship(
    base: Any, local_cls: Any, referred_cls: Any, constraint: Any
): ...
def generate_relationship(
    base: Any,
    direction: Any,
    return_fn: Any,
    attrname: Any,
    local_cls: Any,
    referred_cls: Any,
    **kw: Any,
): ...

class AutomapBase:
    __abstract__: bool = ...
    classes: Any = ...
    @classmethod
    def prepare(
        cls,
        autoload_with: Optional[Any] = ...,
        engine: Optional[Any] = ...,
        reflect: bool = ...,
        schema: Optional[Any] = ...,
        classname_for_table: Optional[Any] = ...,
        collection_class: Optional[Any] = ...,
        name_for_scalar_relationship: Optional[Any] = ...,
        name_for_collection_relationship: Optional[Any] = ...,
        generate_relationship: Optional[Any] = ...,
    ) -> None: ...

def automap_base(declarative_base: Optional[Any] = ..., **kw: Any): ...



# fmt: off
from typing import Any
from typing import Optional

from . import attributes as attributes
from . import interfaces as interfaces
from . import loading as loading
from .interfaces import ORMColumnsClauseRole as ORMColumnsClauseRole
from .path_registry import PathRegistry as PathRegistry
from .util import aliased as aliased
from .util import Bundle as Bundle
from .util import ORMAdapter as ORMAdapter
from .. import future as future
from .. import inspect as inspect
from .. import sql as sql
from .. import util as util
from ..sql import coercions as coercions
from ..sql import expression as expression
from ..sql import roles as roles
from ..sql import visitors as visitors
from ..sql.base import CacheableOptions as CacheableOptions
from ..sql.base import CompileState as CompileState
from ..sql.base import Options as Options
from ..sql.selectable import LABEL_STYLE_DISAMBIGUATE_ONLY as LABEL_STYLE_DISAMBIGUATE_ONLY
from ..sql.selectable import LABEL_STYLE_NONE as LABEL_STYLE_NONE
from ..sql.selectable import LABEL_STYLE_TABLENAME_PLUS_COL as LABEL_STYLE_TABLENAME_PLUS_COL
from ..sql.selectable import SelectState as SelectState
from ..sql.visitors import ExtendedInternalTraversal as ExtendedInternalTraversal
from ..sql.visitors import InternalTraversal as InternalTraversal
# fmt: on

LABEL_STYLE_LEGACY_ORM: Any

class QueryContext:
    class default_load_options(Options): ...
    load_options: Any = ...
    execution_options: Any = ...
    bind_arguments: Any = ...
    compile_state: Any = ...
    query: Any = ...
    session: Any = ...
    loaders_require_buffering: bool = ...
    loaders_require_uniquing: bool = ...
    params: Any = ...
    propagated_loader_options: Any = ...
    attributes: Any = ...
    autoflush: Any = ...
    populate_existing: Any = ...
    invoke_all_eagers: Any = ...
    version_check: Any = ...
    refresh_state: Any = ...
    yield_per: Any = ...
    identity_token: Any = ...
    def __init__(
        self,
        compile_state: Any,
        statement: Any,
        params: Any,
        session: Any,
        load_options: Any,
        execution_options: Optional[Any] = ...,
        bind_arguments: Optional[Any] = ...,
    ) -> None: ...

class ORMCompileState(CompileState):
    class default_compile_options(CacheableOptions): ...
    current_path: Any = ...
    def __init__(self, *arg: Any, **kw: Any) -> None: ...
    @classmethod
    def create_for_statement(
        cls, statement_container: Any, compiler: Any, **kw: Any
    ) -> None: ...
    @classmethod
    def get_column_descriptions(cls, statement: Any): ...
    @classmethod
    def orm_pre_session_exec(
        cls,
        session: Any,
        statement: Any,
        params: Any,
        execution_options: Any,
        bind_arguments: Any,
        is_reentrant_invoke: Any,
    ): ...
    @classmethod
    def orm_setup_cursor_result(
        cls,
        session: Any,
        statement: Any,
        params: Any,
        execution_options: Any,
        bind_arguments: Any,
        result: Any,
    ): ...

class ORMFromStatementCompileState(ORMCompileState):
    multi_row_eager_loaders: bool = ...
    compound_eager_adapter: Any = ...
    extra_criteria_entities: Any = ...
    eager_joins: Any = ...
    use_legacy_query_style: Any = ...
    statement_container: Any = ...
    requested_statement: Any = ...
    dml_table: Any = ...
    compile_options: Any = ...
    statement: Any = ...
    current_path: Any = ...
    attributes: Any = ...
    global_attributes: Any = ...
    primary_columns: Any = ...
    secondary_columns: Any = ...
    create_eager_joins: Any = ...
    order_by: Any = ...
    @classmethod
    def create_for_statement(
        cls, statement_container: Any, compiler: Any, **kw: Any
    ): ...

class ORMSelectCompileState(ORMCompileState, SelectState):
    multi_row_eager_loaders: bool = ...
    compound_eager_adapter: Any = ...
    correlate: Any = ...
    correlate_except: Any = ...
    global_attributes: Any = ...
    select_statement: Any = ...
    for_statement: Any = ...
    use_legacy_query_style: Any = ...
    compile_options: Any = ...
    label_style: Any = ...
    current_path: Any = ...
    eager_order_by: Any = ...
    attributes: Any = ...
    primary_columns: Any = ...
    secondary_columns: Any = ...
    eager_joins: Any = ...
    extra_criteria_entities: Any = ...
    create_eager_joins: Any = ...
    from_clauses: Any = ...
    @classmethod
    def create_for_statement(
        cls, statement: Any, compiler: Any, **kw: Any
    ): ...
    @classmethod
    def determine_last_joined_entity(cls, statement: Any): ...
    @classmethod
    def exported_columns_iterator(cls, statement: Any) -> None: ...
    @classmethod
    def from_statement(cls, statement: Any, from_statement: Any): ...

class _QueryEntity:
    @classmethod
    def to_compile_state(cls, compile_state: Any, entities: Any) -> None: ...

class _MapperEntity(_QueryEntity):
    expr: Any = ...
    mapper: Any = ...
    is_aliased_class: Any = ...
    path: Any = ...
    selectable: Any = ...
    def __init__(self, compile_state: Any, entity: Any) -> None: ...
    supports_single_entity: bool = ...
    use_id_for_hash: bool = ...
    @property
    def type(self): ...
    @property
    def entity_zero_or_selectable(self): ...
    def corresponds_to(self, entity: Any): ...
    def row_processor(self, context: Any, result: Any): ...
    def setup_compile_state(self, compile_state: Any) -> None: ...

class _BundleEntity(_QueryEntity):
    use_id_for_hash: bool = ...
    bundle: Any = ...
    type: Any = ...
    supports_single_entity: Any = ...
    def __init__(
        self,
        compile_state: Any,
        expr: Any,
        setup_entities: bool = ...,
        parent_bundle: Optional[Any] = ...,
    ) -> None: ...
    @property
    def mapper(self): ...
    @property
    def entity_zero(self): ...
    def corresponds_to(self, entity: Any): ...
    @property
    def entity_zero_or_selectable(self): ...
    def setup_compile_state(self, compile_state: Any) -> None: ...
    def row_processor(self, context: Any, result: Any): ...

class _ColumnEntity(_QueryEntity):
    @property
    def type(self): ...
    @property
    def use_id_for_hash(self): ...
    def row_processor(self, context: Any, result: Any): ...

class _RawColumnEntity(_ColumnEntity):
    entity_zero: Any = ...
    mapper: Any = ...
    supports_single_entity: bool = ...
    expr: Any = ...
    column: Any = ...
    entity_zero_or_selectable: Any = ...
    def __init__(
        self,
        compile_state: Any,
        column: Any,
        parent_bundle: Optional[Any] = ...,
    ) -> None: ...
    def corresponds_to(self, entity: Any): ...
    def setup_compile_state(self, compile_state: Any) -> None: ...

class _ORMColumnEntity(_ColumnEntity):
    supports_single_entity: bool = ...
    expr: Any = ...
    entity_zero: Any = ...
    mapper: Any = ...
    column: Any = ...
    def __init__(
        self,
        compile_state: Any,
        column: Any,
        parententity: Any,
        parent_bundle: Optional[Any] = ...,
    ) -> None: ...
    def corresponds_to(self, entity: Any): ...
    def setup_compile_state(self, compile_state: Any) -> None: ...

class _IdentityTokenEntity(_ORMColumnEntity):
    def setup_compile_state(self, compile_state: Any) -> None: ...
    def row_processor(self, context: Any, result: Any): ...



import sys
from typing import Any
from typing import Callable
from typing import Collection
from typing import ContextManager
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import overload
from typing import Sequence
from typing import Type
from typing import TypeVar
from typing import Union

from typing_extensions import Literal
from typing_extensions import Protocol

from .identity import WeakInstanceDict
from .mapper import Mapper
from .path_registry import PathRegistry
from .query import Query
from .state import InstanceState
from .. import util
from ..engine import Connection
from ..engine import Engine
from ..engine import Result
from ..engine.base import _ExecutionOptions
from ..sql import ClauseElement
from ..sql import Executable
from ..sql.base import Options

_T = TypeVar("_T")
_TSession = TypeVar("_TSession", bound=Session)
_TSessionTransaction = TypeVar(
    "_TSessionTransaction", bound=SessionTransaction
)

_TSharedSessionProtocol = TypeVar(
    "_TSharedSessionProtocol", bound=_SharedSessionProtocol[Any]
)
_TSessionTransactionProtocol = TypeVar(
    "_TSessionTransactionProtocol", bound=_SessionTransactionProtocol
)
_TSessionProtocol = TypeVar("_TSessionProtocol", bound=_SessionProtocol)

_BindArguments = Mapping[str, Any]

class _IdentityMap:
    @overload
    def __get__(self, instance: None, owner: Any) -> None: ...
    @overload
    def __get__(self, instance: object, owner: Any) -> WeakInstanceDict: ...

class _SharedSessionProtocol(Protocol[_T]):
    identity_map: _IdentityMap
    autoflush: bool
    @property
    def dirty(self) -> util.IdentitySet[Any]: ...
    @property
    def deleted(self) -> util.IdentitySet[Any]: ...
    @property
    def new(self) -> util.IdentitySet[Any]: ...
    @property
    def is_active(self) -> bool: ...
    @property
    def no_autoflush(
        self: _TSharedSessionProtocol,
    ) -> ContextManager[_TSharedSessionProtocol]: ...
    @util.memoized_property
    def info(self) -> Mapping[Any, Any]: ...
    def __contains__(self, instance: Any) -> bool: ...
    def __iter__(self) -> Iterator[Any]: ...
    def add(self, instance: Any, _warn: bool = ...) -> None: ...
    def add_all(self, instances: Iterable[Any]) -> None: ...
    def expire_all(self) -> None: ...
    def expire(
        self, instance: Any, attribute_names: Optional[Iterable[str]] = ...
    ) -> None: ...
    def expunge(self, instance: Any) -> None: ...
    def expunge_all(self) -> None: ...
    def get_bind(
        self,
        mapper: Optional[Any] = ...,
        clause: Optional[ClauseElement] = ...,
        bind: Optional[_T] = ...,
        _sa_skip_events: Optional[Any] = ...,
        _sa_skip_for_implicit_returning: bool = ...,
    ) -> _T: ...
    def is_modified(
        self, instance: Any, include_collections: bool = ...
    ) -> bool: ...
    def in_transaction(self) -> bool: ...

class _SessionProtocol(
    _SharedSessionProtocol[Union[Connection, Engine]], Protocol
):
    def __enter__(self: _TSessionProtocol) -> _TSessionProtocol: ...
    def __exit__(self, type_: Any, value: Any, traceback: Any) -> None: ...
    def begin(
        self,
        subtransactions: bool = ...,  # NOTE: Deprecated.
        nested: bool = ...,
    ) -> _SessionTransactionProtocol: ...
    def begin_nested(self) -> _SessionTransactionProtocol: ...
    def rollback(self) -> None: ...
    def commit(self) -> None: ...
    def connection(self) -> Connection: ...
    def execute(
        self,
        statement: Executable,
        params: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
        ] = ...,
        execution_options: _ExecutionOptions = ...,
        bind_arguments: Optional[_BindArguments] = ...,
        **kw: Any,
    ) -> Result: ...
    def scalar(
        self,
        statement: Executable,
        params: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
        ] = ...,
        execution_options: _ExecutionOptions = ...,
        bind_arguments: Optional[_BindArguments] = ...,
        **kw: Any,
    ) -> Any: ...
    def close(self) -> None: ...
    def refresh(
        self,
        instance: Any,
        attribute_names: Optional[Iterable[str]] = ...,
        with_for_update: Optional[
            Union[Literal[True], Mapping[str, Any]]
        ] = ...,
    ) -> None: ...
    def delete(self, instance: Any) -> None: ...
    def get(
        self,
        entity: Any,
        ident: Any,
        options: Optional[Sequence[Any]] = ...,
        populate_existing: bool = ...,
        with_for_update: Optional[
            Union[Literal[True], Mapping[str, Any]]
        ] = ...,
        identity_token: Optional[Any] = ...,
    ) -> Any: ...
    def merge(self, instance: _T, load: bool = ...) -> _T: ...
    def flush(self, objects: Optional[Collection[Any]] = ...) -> None: ...
    @classmethod
    def close_all(cls) -> None: ...  # NOTE: Deprecated.

class _SessionTransactionProtocol(Protocol):
    nested: bool
    @property
    def is_active(self) -> bool: ...
    @property
    def parent(
        self,
    ) -> Optional[_SessionTransactionProtocol]: ...
    def connection(
        self,
        bindkey: Any,
        execution_options: Optional[_ExecutionOptions] = ...,
        **kwargs: Any,
    ) -> Connection: ...
    def prepare(self) -> None: ...
    def commit(self) -> Optional[_SessionTransactionProtocol]: ...
    def rollback(self) -> Optional[_SessionTransactionProtocol]: ...
    def close(self, invalidate: bool = ...) -> None: ...
    def __enter__(
        self: _TSessionTransactionProtocol,
    ) -> _TSessionTransactionProtocol: ...
    def __exit__(self, type_: Any, value: Any, traceback: Any) -> None: ...

if sys.version_info >= (3, 0):
    from typing import Generator
    from ..ext.asyncio import AsyncEngine, AsyncConnection

    _TSessionMakerType = TypeVar(
        "_TSessionMakerType",
        bound=Union[_SessionProtocol, _AsyncSessionProtocol],
    )
    _TAsyncSessionTransactionProtocol = TypeVar(
        "_TAsyncSessionTransactionProtocol",
        bound=_AsyncSessionTransactionProtocol,
    )
    _TAsyncSessionProtocol = TypeVar(
        "_TAsyncSessionProtocol", bound=_AsyncSessionProtocol
    )
    class _AsyncSessionTransactionProtocol(Protocol):
        @property
        def is_active(self) -> bool: ...
        async def commit(
            self,
        ) -> Optional[_AsyncSessionTransactionProtocol]: ...
        async def rollback(
            self,
        ) -> Optional[_AsyncSessionTransactionProtocol]: ...
        async def start(
            self: _TAsyncSessionTransactionProtocol,
        ) -> _TAsyncSessionTransactionProtocol: ...
        def __await__(
            self: _TAsyncSessionTransactionProtocol,
        ) -> Generator[Any, None, _TAsyncSessionTransactionProtocol]: ...
        async def __aenter__(
            self: _TAsyncSessionTransactionProtocol,
        ) -> _TAsyncSessionTransactionProtocol: ...
        async def __aexit__(
            self, type_: Any, value: Any, traceback: Any
        ) -> None: ...
    class _AsyncSessionProtocol(
        _SharedSessionProtocol[Union[AsyncConnection, AsyncEngine]], Protocol
    ):
        async def __aenter__(
            self: _TAsyncSessionProtocol,
        ) -> _TAsyncSessionProtocol: ...
        async def __aexit__(
            self, type_: Any, value: Any, traceback: Any
        ) -> None: ...
        def begin(self) -> _AsyncSessionTransactionProtocol: ...
        def begin_nested(self) -> _AsyncSessionTransactionProtocol: ...
        async def rollback(self) -> None: ...
        async def commit(self) -> None: ...
        async def connection(self) -> Any: ...
        async def execute(
            self,
            statement: Executable,
            params: Optional[
                Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
            ] = ...,
            execution_options: _ExecutionOptions = ...,
            bind_arguments: Optional[_BindArguments] = ...,
            **kw: Any,
        ) -> Result: ...
        async def scalar(
            self,
            statement: Executable,
            params: Optional[
                Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
            ] = ...,
            execution_options: _ExecutionOptions = ...,
            bind_arguments: Optional[_BindArguments] = ...,
            **kw: Any,
        ) -> Any: ...
        async def close(self) -> None: ...
        async def refresh(
            self,
            instance: Any,
            attribute_names: Optional[Iterable[str]] = ...,
            with_for_update: Optional[
                Union[Literal[True], Mapping[str, Any]]
            ] = ...,
        ) -> None: ...
        async def get(
            self,
            entity: Any,
            ident: Any,
            options: Optional[Sequence[Any]] = ...,
            populate_existing: bool = ...,
            with_for_update: Optional[
                Union[Literal[True], Mapping[str, Any]]
            ] = ...,
            identity_token: Optional[Any] = ...,
        ) -> Any: ...
        async def stream(
            self,
            statement: Any,
            params: Optional[
                Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
            ] = ...,
            execution_options: _ExecutionOptions = ...,
            bind_arguments: Optional[_BindArguments] = ...,
            **kw: Any,
        ) -> Any: ...
        async def delete(self, instance: Any) -> None: ...
        async def merge(self, instance: _T, load: bool = ...) -> _T: ...
        async def flush(
            self, objects: Optional[Collection[Any]] = ...
        ) -> None: ...
        @classmethod
        async def close_all(cls) -> None: ...  # NOTE: Deprecated.

else:
    _TSessionMakerType = TypeVar("_TSessionMakerType", bound=_SessionProtocol)

class _SessionClassMethods:
    @classmethod
    def close_all(cls) -> None: ...  # NOTE: Deprecated.
    @classmethod
    def identity_key(cls, *args: Any, **kwargs: Any) -> Any: ...
    @classmethod
    def object_session(cls, instance: Any) -> Session: ...

ACTIVE: util.langhelpers._symbol
PREPARED: util.langhelpers._symbol
COMMITTED: util.langhelpers._symbol
DEACTIVE: util.langhelpers._symbol
CLOSED: util.langhelpers._symbol

class ORMExecuteState(util.MemoizedSlots):
    session: Session
    statement: Executable
    parameters: Mapping[str, Any]
    local_execution_options: Any
    execution_options: Any
    bind_arguments: _BindArguments
    def __init__(
        self,
        session: Session,
        statement: Executable,
        parameters: Mapping[str, Any],
        execution_options: _ExecutionOptions,
        bind_arguments: _BindArguments,
        compile_state_cls: Any,
        events_todo: Any,
    ) -> None: ...
    def invoke_statement(
        self,
        statement: Optional[Any] = ...,
        params: Optional[Mapping[str, Any]] = ...,
        execution_options: _ExecutionOptions = ...,
        bind_arguments: Optional[_BindArguments] = ...,
    ) -> Any: ...
    @property
    def bind_mapper(self) -> Mapper: ...
    @property
    def all_mappers(self) -> List[Mapper]: ...
    @property
    def is_orm_statement(self) -> bool: ...
    @property
    def is_select(self) -> bool: ...
    @property
    def is_insert(self) -> bool: ...
    @property
    def is_update(self) -> bool: ...
    @property
    def is_delete(self) -> bool: ...
    def update_execution_options(self, **opts: Any) -> None: ...
    @property
    def lazy_loaded_from(self) -> InstanceState: ...
    @property
    def loader_strategy_path(self) -> Optional[PathRegistry]: ...
    @property
    def is_column_load(self) -> bool: ...
    @property
    def is_relationship_load(self) -> bool: ...
    @property
    def load_options(self) -> Options: ...
    @property
    def update_delete_options(self) -> Options: ...
    @property
    def user_defined_options(self) -> List[Options]: ...

class SessionTransaction:
    session: Session
    nested: bool
    def __init__(
        self,
        session: Session,
        parent: Optional[SessionTransaction] = ...,
        nested: bool = ...,
        autobegin: bool = ...,
    ) -> None: ...
    @property
    def parent(self) -> SessionTransaction: ...
    @property
    def is_active(self) -> bool: ...
    def connection(
        self,
        bindkey: Any,
        execution_options: Optional[_ExecutionOptions] = ...,
        **kwargs: Any,
    ) -> Connection: ...
    def prepare(self) -> None: ...
    def commit(self, _to_root: bool = ...) -> Optional[SessionTransaction]: ...
    def rollback(
        self, _capture_exception: bool = ..., _to_root: bool = ...
    ) -> Optional[SessionTransaction]: ...
    def close(self, invalidate: bool = ...) -> None: ...
    def __enter__(self: _TSessionTransaction) -> _TSessionTransaction: ...
    def __exit__(self, type_: Any, value: Any, traceback: Any) -> None: ...

class Session(_SessionClassMethods):
    identity_map: _IdentityMap
    bind: Optional[Union[Connection, Engine]]
    future: bool
    hash_key: int
    autoflush: bool
    expire_on_commit: bool
    enable_baked_queries: bool
    autocommit: bool
    twophase: bool
    def __init__(
        self,
        bind: Optional[Union[Connection, Engine]] = ...,
        autoflush: bool = ...,
        future: bool = ...,
        expire_on_commit: bool = ...,
        autocommit: bool = ...,  # NOTE: Deprecated.
        twophase: bool = ...,
        binds: Optional[Mapping[Any, Union[Connection, Engine]]] = ...,
        enable_baked_queries: bool = ...,
        info: Optional[Mapping[Any, Any]] = ...,
        query_cls: Optional[
            Union[Type[Query[Any]], Callable[..., Query[Any]]]
        ] = ...,
    ) -> None: ...
    connection_callable: Any = ...
    def __enter__(self: _TSession) -> _TSession: ...
    def __exit__(self, type_: Any, value: Any, traceback: Any) -> None: ...
    @property
    def transaction(
        self,
    ) -> Optional[SessionTransaction]: ...  # NOTE: Deprecated.
    def in_transaction(self) -> bool: ...
    def in_nested_transaction(self) -> bool: ...
    def get_transaction(self) -> Optional[SessionTransaction]: ...
    def get_nested_transaction(self) -> Optional[SessionTransaction]: ...
    @util.memoized_property
    def info(self) -> Mapping[Any, Any]: ...
    def begin(
        self,
        subtransactions: bool = ...,  # NOTE: Deprecated.
        nested: bool = ...,
        _subtrans: bool = ...,
    ) -> SessionTransaction: ...
    def begin_nested(self) -> SessionTransaction: ...
    def rollback(self) -> None: ...
    def commit(self) -> None: ...
    def prepare(self) -> None: ...
    def connection(
        self,
        bind_arguments: Optional[_BindArguments] = ...,
        close_with_result: bool = ...,
        execution_options: _ExecutionOptions = ...,
        **kw: Any,
    ) -> Connection: ...
    def execute(
        self,
        statement: Executable,
        params: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
        ] = ...,
        execution_options: _ExecutionOptions = ...,
        bind_arguments: Optional[_BindArguments] = ...,
        _parent_execute_state: Optional[Any] = ...,
        _add_event: Optional[Any] = ...,
        **kw: Any,
    ) -> Result: ...
    def scalar(
        self,
        statement: Executable,
        params: Optional[
            Union[Mapping[str, Any], Sequence[Mapping[str, Any]]]
        ] = ...,
        execution_options: _ExecutionOptions = ...,
        bind_arguments: Optional[_BindArguments] = ...,
        **kw: Any,
    ) -> Any: ...
    def close(self) -> None: ...
    def invalidate(self) -> None: ...
    def expunge_all(self) -> None: ...
    def bind_mapper(self, mapper: Any, bind: Any) -> None: ...
    def bind_table(self, table: Any, bind: Any) -> None: ...
    def get_bind(
        self,
        mapper: Optional[Any] = ...,
        clause: Optional[ClauseElement] = ...,
        bind: Optional[Union[Connection, Engine]] = ...,
        _sa_skip_events: Optional[Any] = ...,
        _sa_skip_for_implicit_returning: bool = ...,
    ) -> Union[Connection, Engine]: ...
    def query(self, *entities: Any, **kwargs: Any) -> Query[Any]: ...
    @property
    def no_autoflush(self: _TSession) -> ContextManager[_TSession]: ...
    def refresh(
        self,
        instance: Any,
        attribute_names: Optional[Iterable[str]] = ...,
        with_for_update: Optional[
            Union[Literal[True], Mapping[str, Any]]
        ] = ...,
    ) -> None: ...
    def expire_all(self) -> None: ...
    def expire(
        self, instance: Any, attribute_names: Optional[Iterable[str]] = ...
    ) -> None: ...
    def expunge(self, instance: Any) -> None: ...
    def add(self, instance: Any, _warn: bool = ...) -> None: ...
    def add_all(self, instances: Any) -> None: ...
    def delete(self, instance: Any) -> None: ...
    def get(
        self,
        entity: Any,
        ident: Any,
        options: Optional[Sequence[Any]] = ...,
        populate_existing: bool = ...,
        with_for_update: Optional[
            Union[Literal[True], Mapping[str, Any]]
        ] = ...,
        identity_token: Optional[Any] = ...,
    ) -> Any: ...
    def merge(self, instance: _T, load: bool = ...) -> _T: ...
    def enable_relationship_loading(self, obj: Any) -> None: ...
    def __contains__(self, instance: Any) -> bool: ...
    def __iter__(self) -> Iterator[Any]: ...
    def flush(self, objects: Optional[Collection[Any]] = ...) -> None: ...
    def bulk_save_objects(
        self,
        objects: Sequence[Any],
        return_defaults: bool = ...,
        update_changed_only: bool = ...,
        preserve_order: bool = ...,
    ) -> None: ...
    def bulk_insert_mappings(
        self,
        mapper: Any,
        mappings: Sequence[Mapping[str, Any]],
        return_defaults: bool = ...,
        render_nulls: bool = ...,
    ) -> None: ...
    def bulk_update_mappings(
        self, mapper: Any, mappings: Sequence[Mapping[str, Any]]
    ) -> None: ...
    def is_modified(
        self, instance: Any, include_collections: bool = ...
    ) -> bool: ...
    @property
    def is_active(self) -> bool: ...
    @property
    def dirty(self) -> util.IdentitySet[Any]: ...
    @property
    def deleted(self) -> util.IdentitySet[Any]: ...
    @property
    def new(self) -> util.IdentitySet[Any]: ...

class sessionmaker(_SessionClassMethods, Generic[_TSessionMakerType]):
    kw: Mapping[str, Any]
    class_: Type[_TSessionMakerType]
    if sys.version_info >= (3, 0):
        @overload
        def __init__(
            self: sessionmaker[Session],
            bind: Optional[Union[Connection, Engine]] = ...,
            class_: None = ...,
            autoflush: bool = ...,
            autocommit: bool = ...,  # NOTE: Deprecated.
            expire_on_commit: bool = ...,
            info: Optional[Mapping[Any, Any]] = ...,
            **kw: Any,
        ) -> None: ...
        @overload
        def __init__(
            self,
            bind: Optional[
                Union[Connection, Engine, AsyncConnection, AsyncEngine]
            ],
            class_: Type[_TSessionMakerType],
            autoflush: bool = ...,
            autocommit: bool = ...,  # NOTE: Deprecated.
            expire_on_commit: bool = ...,
            info: Optional[Mapping[Any, Any]] = ...,
            **kw: Any,
        ) -> None: ...
        @overload
        def __init__(
            self,
            *,
            bind: Optional[
                Union[Connection, Engine, AsyncConnection, AsyncEngine]
            ] = ...,
            class_: Type[_TSessionMakerType],
            autoflush: bool = ...,
            autocommit: bool = ...,  # NOTE: Deprecated.
            expire_on_commit: bool = ...,
            info: Optional[Mapping[Any, Any]] = ...,
            **kw: Any,
        ) -> None: ...
    else:
        @overload
        def __init__(
            self: sessionmaker[Session],
            bind: Optional[Union[Connection, Engine]] = ...,
            class_: None = ...,
            autoflush: bool = ...,
            autocommit: bool = ...,  # NOTE: Deprecated.
            expire_on_commit: bool = ...,
            info: Optional[Mapping[Any, Any]] = ...,
            **kw: Any,
        ) -> None: ...
        @overload
        def __init__(
            self,
            bind: Optional[Union[Connection, Engine]],
            class_: Type[_TSessionMakerType],
            autoflush: bool = ...,
            autocommit: bool = ...,  # NOTE: Deprecated.
            expire_on_commit: bool = ...,
            info: Optional[Mapping[Any, Any]] = ...,
            **kw: Any,
        ) -> None: ...
        @overload
        def __init__(
            self,
            *,
            bind: Optional[Union[Connection, Engine]] = ...,
            class_: Type[_TSessionMakerType],
            autoflush: bool = ...,
            autocommit: bool = ...,  # NOTE: Deprecated.
            expire_on_commit: bool = ...,
            info: Optional[Mapping[Any, Any]] = ...,
            **kw: Any,
        ) -> None: ...
    # NOTE: The return type of `begin()` isn't technically correct, but since
    # `Session.__enter__` and `AsyncSession.__aenter__` both return self,
    # returning `_TSessionMakerType` from `begin()` works out mostly
    # structurally the same as what actually takes place in `sessionmaker.begin()`
    def begin(self) -> _TSessionMakerType: ...
    def __call__(self, **local_kw: Any) -> _TSessionMakerType: ...
    def configure(self, **new_kw: Any) -> None: ...

def close_all_sessions() -> None: ...
def make_transient(instance: object) -> None: ...
def make_transient_to_detached(instance: object) -> None: ...

if sys.version_info >= (3, 0):
    def object_session(
        instance: object,
    ) -> Optional[Union[Session, _SessionProtocol, _AsyncSessionProtocol]]: ...

else:
    def object_session(instance: object) -> Session: ...



# fmt: off
from .base import Pool as Pool
from .base import reset_commit as reset_commit
from .base import reset_none as reset_none
from .base import reset_rollback as reset_rollback
from .dbapi_proxy import clear_managers as clear_managers
from .dbapi_proxy import manage as manage
from .impl import AssertionPool as AssertionPool
from .impl import AsyncAdaptedQueuePool as AsyncAdaptedQueuePool
from .impl import FallbackAsyncAdaptedQueuePool as FallbackAsyncAdaptedQueuePool
from .impl import NullPool as NullPool
from .impl import QueuePool as QueuePool
from .impl import SingletonThreadPool as SingletonThreadPool
from .impl import StaticPool as StaticPool



from typing import Any
from typing import TypeVar

from . import elements
from . import type_api
from ..util import langhelpers

_TE = TypeVar("_TE", bound=type_api.TypeEngine[Any])

REQUIRED: langhelpers._symbol

class _multiparam_column(elements.ColumnElement[_TE]):
    index: Any = ...
    key: str = ...
    original: elements.ColumnElement[_TE] = ...
    default: Any = ...
    type: _TE = ...  # type: ignore[assignment]
    def __init__(
        self, original: elements.ColumnElement[_TE], index: Any
    ) -> None: ...
    def compare(self, other: Any, **kw: Any) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...  # type: ignore[override]



from typing import Any
from typing import Dict
from typing import Tuple
from typing import Type
from typing import TypeVar

from .. import util

_A = TypeVar("_A", bound=Annotated)
_T = TypeVar("_T")

EMPTY_ANNOTATIONS: util.immutabledict[None, None]

class SupportsAnnotations: ...
class SupportsCloneAnnotations(SupportsAnnotations): ...
class SupportsWrappingAnnotations(SupportsAnnotations): ...

class Annotated:
    def __new__(cls: Type[_T], *args: Any) -> _T: ...
    __dict__: Dict[Any, Any] = ...
    def __init__(self, element: Any, values: Any) -> None: ...
    def __reduce__(self: _A) -> Tuple[Type[_A], Tuple[Any, Any]]: ...
    def __hash__(self) -> int: ...
    def __eq__(self, other: Any) -> bool: ...
    @property
    def entity_namespace(self) -> Any: ...

annotated_classes: Dict[Type[Any], Any]



from collections import namedtuple
from typing import Any
from typing import Generic
from typing import Optional
from typing import overload
from typing import Type
from typing import TypeVar

from . import collections as collections
from . import interfaces as interfaces
from .base import ATTR_EMPTY as ATTR_EMPTY
from .base import ATTR_WAS_SET as ATTR_WAS_SET
from .base import CALLABLES_OK as CALLABLES_OK
from .base import INIT_OK as INIT_OK
from .base import instance_dict as instance_dict
from .base import instance_state as instance_state
from .base import instance_str as instance_str
from .base import LOAD_AGAINST_COMMITTED as LOAD_AGAINST_COMMITTED
from .base import manager_of_class as manager_of_class
from .base import NEVER_SET as NEVER_SET
from .base import NO_AUTOFLUSH as NO_AUTOFLUSH
from .base import NO_CHANGE as NO_CHANGE
from .base import NO_RAISE as NO_RAISE
from .base import NO_VALUE as NO_VALUE
from .base import NON_PERSISTENT_OK as NON_PERSISTENT_OK
from .base import PASSIVE_CLASS_MISMATCH as PASSIVE_CLASS_MISMATCH
from .base import PASSIVE_NO_FETCH as PASSIVE_NO_FETCH
from .base import PASSIVE_NO_FETCH_RELATED as PASSIVE_NO_FETCH_RELATED
from .base import PASSIVE_NO_INITIALIZE as PASSIVE_NO_INITIALIZE
from .base import PASSIVE_NO_RESULT as PASSIVE_NO_RESULT
from .base import PASSIVE_OFF as PASSIVE_OFF
from .base import PASSIVE_ONLY_PERSISTENT as PASSIVE_ONLY_PERSISTENT
from .base import PASSIVE_RETURN_NO_VALUE as PASSIVE_RETURN_NO_VALUE
from .base import RELATED_OBJECT_OK as RELATED_OBJECT_OK
from .base import SQL_OK as SQL_OK
from .base import state_str as state_str
from .. import event as event
from .. import inspection as inspection
from .. import util as util
from ..sql import base as sql_base
from ..sql import ClauseElement
from ..sql import roles as roles
from ..sql import traversals as traversals
from ..sql import visitors as visitors

class QueryableAttribute(
    interfaces._MappedAttribute,
    interfaces.InspectionAttr,
    interfaces.PropComparator,
    traversals.HasCopyInternals,
    roles.JoinTargetRole,
    roles.OnClauseRole,
    sql_base.Immutable,
    sql_base.MemoizedHasCacheKey,
):
    is_attribute: bool = ...
    class_: Any = ...
    key: Any = ...
    impl: Any = ...
    comparator: Any = ...
    def __init__(
        self,
        class_: Any,
        key: Any,
        parententity: Any,
        impl: Optional[Any] = ...,
        comparator: Optional[Any] = ...,
        of_type: Optional[Any] = ...,
        extra_criteria: Any = ...,
    ) -> None: ...
    def __reduce__(self): ...
    def get_history(self, instance: Any, passive: Any = ...): ...
    def info(self): ...
    def parent(self): ...
    def expression(self): ...
    def __clause_element__(self): ...
    def adapt_to_entity(self, adapt_to_entity: Any): ...
    def of_type(self, entity: Any): ...
    def and_(self, *other: Any): ...
    def label(self, name: Any): ...
    def operate(self, op: Any, *other: Any, **kwargs: Any): ...
    def reverse_operate(self, op: Any, other: Any, **kwargs: Any): ...
    def hasparent(self, state: Any, optimistic: bool = ...): ...
    def __getattr__(self, key: Any): ...
    def property(self): ...

_T = TypeVar("_T")

class Mapped(QueryableAttribute, Generic[_T]):
    def __init__(self, type: Type[_T]) -> None: ...
    @overload
    def __get__(self, instance: None, owner: Any) -> "Mapped"[_T]: ...
    @overload
    def __get__(self, instance: object, owner: Any) -> _T: ...
    @overload
    def __set__(self, instance: Any, value: _T) -> None: ...
    @overload
    def __set__(self, instance: Any, value: ClauseElement) -> None: ...
    def __delete__(self, instance: Any) -> None: ...
    @classmethod
    def _empty_constructor(cls, arg1: Any) -> "Mapped"[_T]: ...

class InstrumentedAttribute(Mapped):
    inherit_cache: bool = ...

HasEntityNamespace = namedtuple("HasEntityNamespace", ["entity_namespace"])

def create_proxied_attribute(descriptor: Any): ...

OP_REMOVE: Any
OP_APPEND: Any
OP_REPLACE: Any
OP_BULK_REPLACE: Any
OP_MODIFIED: Any

class AttributeEvent:
    impl: Any = ...
    op: Any = ...
    parent_token: Any = ...
    def __init__(self, attribute_impl: Any, op: Any) -> None: ...
    def __eq__(self, other: Any) -> Any: ...
    @property
    def key(self): ...
    def hasparent(self, state: Any): ...

Event = AttributeEvent

class AttributeImpl:
    class_: Any = ...
    key: Any = ...
    callable_: Any = ...
    dispatch: Any = ...
    trackparent: Any = ...
    parent_token: Any = ...
    send_modified_events: Any = ...
    is_equal: Any = ...
    accepts_scalar_loader: Any = ...
    load_on_unexpire: Any = ...
    def __init__(
        self,
        class_: Any,
        key: Any,
        callable_: Any,
        dispatch: Any,
        trackparent: bool = ...,
        compare_function: Optional[Any] = ...,
        active_history: bool = ...,
        parent_token: Optional[Any] = ...,
        load_on_unexpire: bool = ...,
        send_modified_events: bool = ...,
        accepts_scalar_loader: Optional[Any] = ...,
        **kwargs: Any,
    ) -> None: ...
    active_history: Any = ...
    def hasparent(self, state: Any, optimistic: bool = ...): ...
    def sethasparent(
        self, state: Any, parent_state: Any, value: Any
    ) -> None: ...
    def get_history(
        self, state: Any, dict_: Any, passive: Any = ...
    ) -> None: ...
    def get_all_pending(
        self, state: Any, dict_: Any, passive: Any = ...
    ) -> None: ...
    def get(self, state: Any, dict_: Any, passive: Any = ...): ...
    def append(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
    ) -> None: ...
    def remove(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
    ) -> None: ...
    def pop(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
    ) -> None: ...
    def set(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
        check_old: Optional[Any] = ...,
        pop: bool = ...,
    ) -> None: ...
    def get_committed_value(
        self, state: Any, dict_: Any, passive: Any = ...
    ): ...
    def set_committed_value(self, state: Any, dict_: Any, value: Any): ...

class ScalarAttributeImpl(AttributeImpl):
    default_accepts_scalar_loader: bool = ...
    uses_objects: bool = ...
    supports_population: bool = ...
    collection: bool = ...
    dynamic: bool = ...
    def __init__(self, *arg: Any, **kw: Any) -> None: ...
    def delete(self, state: Any, dict_: Any) -> None: ...
    def get_history(self, state: Any, dict_: Any, passive: Any = ...): ...
    def set(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
        check_old: Optional[Any] = ...,
        pop: bool = ...,
    ) -> None: ...
    def fire_replace_event(
        self, state: Any, dict_: Any, value: Any, previous: Any, initiator: Any
    ): ...
    def fire_remove_event(
        self, state: Any, dict_: Any, value: Any, initiator: Any
    ) -> None: ...
    @property
    def type(self) -> None: ...

class ScalarObjectAttributeImpl(ScalarAttributeImpl):
    default_accepts_scalar_loader: bool = ...
    uses_objects: bool = ...
    supports_population: bool = ...
    collection: bool = ...
    def delete(self, state: Any, dict_: Any) -> None: ...
    def get_history(self, state: Any, dict_: Any, passive: Any = ...): ...
    def get_all_pending(self, state: Any, dict_: Any, passive: Any = ...): ...
    def set(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
        check_old: Optional[Any] = ...,
        pop: bool = ...,
    ) -> None: ...
    def fire_remove_event(
        self, state: Any, dict_: Any, value: Any, initiator: Any
    ) -> None: ...
    def fire_replace_event(
        self, state: Any, dict_: Any, value: Any, previous: Any, initiator: Any
    ): ...

class CollectionAttributeImpl(AttributeImpl):
    default_accepts_scalar_loader: bool = ...
    uses_objects: bool = ...
    supports_population: bool = ...
    collection: bool = ...
    dynamic: bool = ...
    copy: Any = ...
    collection_factory: Any = ...
    def __init__(
        self,
        class_: Any,
        key: Any,
        callable_: Any,
        dispatch: Any,
        typecallable: Optional[Any] = ...,
        trackparent: bool = ...,
        copy_function: Optional[Any] = ...,
        compare_function: Optional[Any] = ...,
        **kwargs: Any,
    ) -> None: ...
    def get_history(self, state: Any, dict_: Any, passive: Any = ...): ...
    def get_all_pending(self, state: Any, dict_: Any, passive: Any = ...): ...
    def fire_append_event(
        self, state: Any, dict_: Any, value: Any, initiator: Any
    ): ...
    def fire_pre_remove_event(
        self, state: Any, dict_: Any, initiator: Any
    ) -> None: ...
    def fire_remove_event(
        self, state: Any, dict_: Any, value: Any, initiator: Any
    ) -> None: ...
    def delete(self, state: Any, dict_: Any) -> None: ...
    def append(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
    ) -> None: ...
    def remove(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
    ) -> None: ...
    def pop(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Any,
        passive: Any = ...,
    ) -> None: ...
    def set(
        self,
        state: Any,
        dict_: Any,
        value: Any,
        initiator: Optional[Any] = ...,
        passive: Any = ...,
        check_old: Optional[Any] = ...,
        pop: bool = ...,
    ) -> None: ...
    def set_committed_value(self, state: Any, dict_: Any, value: Any): ...
    def get_collection(
        self,
        state: Any,
        dict_: Any,
        user_data: Optional[Any] = ...,
        passive: Any = ...,
    ): ...

def backref_listeners(attribute: Any, key: Any, uselist: Any): ...

class History:
    def __bool__(self): ...
    __nonzero__: Any = ...
    def empty(self): ...
    def sum(self): ...
    def non_deleted(self): ...
    def non_added(self): ...
    def has_changes(self): ...
    def as_state(self): ...
    @classmethod
    def from_scalar_attribute(
        cls, attribute: Any, state: Any, current: Any
    ): ...
    @classmethod
    def from_object_attribute(
        cls, attribute: Any, state: Any, current: Any
    ): ...
    @classmethod
    def from_collection(cls, attribute: Any, state: Any, current: Any): ...

HISTORY_BLANK: Any

def get_history(obj: Any, key: Any, passive: Any = ...): ...
def get_state_history(state: Any, key: Any, passive: Any = ...): ...
def has_parent(cls, obj: Any, key: Any, optimistic: bool = ...): ...
def register_attribute(class_: Any, key: Any, **kw: Any): ...
def register_attribute_impl(
    class_: Any,
    key: Any,
    uselist: bool = ...,
    callable_: Optional[Any] = ...,
    useobject: bool = ...,
    impl_class: Optional[Any] = ...,
    backref: Optional[Any] = ...,
    **kw: Any,
): ...
def register_descriptor(
    class_: Any,
    key: Any,
    comparator: Optional[Any] = ...,
    parententity: Optional[Any] = ...,
    doc: Optional[Any] = ...,
): ...
def unregister_attribute(class_: Any, key: Any) -> None: ...
def init_collection(obj: Any, key: Any): ...
def init_state_collection(state: Any, dict_: Any, key: Any): ...
def set_committed_value(instance: Any, key: Any, value: Any) -> None: ...
def set_attribute(
    instance: Any, key: Any, value: Any, initiator: Optional[Any] = ...
) -> None: ...
def get_attribute(instance: Any, key: Any): ...
def del_attribute(instance: Any, key: Any) -> None: ...
def flag_modified(instance: Any, key: Any) -> None: ...
def flag_dirty(instance: Any) -> None: ...



# fmt: off
from .base import ColumnCollection as ColumnCollection
from .dml import Delete as Delete
from .dml import Insert as Insert
from .dml import Update as Update
from .elements import between as between
from .elements import BindParameter
from .elements import BooleanClauseList
from .elements import Case
from .elements import Cast
from .elements import ClauseElement as ClauseElement
from .elements import collate as collate
from .elements import CollectionAggregate
from .elements import ColumnClause
from .elements import ColumnElement as ColumnElement
from .elements import Extract
from .elements import False_ as False_
from .elements import FunctionFilter
from .elements import Label
from .elements import literal as literal
from .elements import literal_column as literal_column
from .elements import not_ as not_
from .elements import Null
from .elements import outparam as outparam
from .elements import Over
from .elements import quoted_name as quoted_name
from .elements import TextClause
from .elements import True_ as True_
from .elements import Tuple
from .elements import TypeCoerce
from .elements import UnaryExpression
from .elements import WithinGroup
from .functions import func as func
from .functions import modifier as modifier
from .lambdas import lambda_stmt as lambda_stmt
from .lambdas import LambdaElement as LambdaElement
from .lambdas import StatementLambdaElement as StatementLambdaElement
from .operators import custom_op as custom_op
from .selectable import Alias as Alias
from .selectable import AliasedReturnsRows as AliasedReturnsRows
from .selectable import CompoundSelect as CompoundSelect
from .selectable import CTE
from .selectable import Exists
from .selectable import FromClause as FromClause
from .selectable import Join as Join
from .selectable import LABEL_STYLE_DEFAULT as LABEL_STYLE_DEFAULT
from .selectable import LABEL_STYLE_DISAMBIGUATE_ONLY as LABEL_STYLE_DISAMBIGUATE_ONLY
from .selectable import LABEL_STYLE_NONE as LABEL_STYLE_NONE
from .selectable import LABEL_STYLE_TABLENAME_PLUS_COL as LABEL_STYLE_TABLENAME_PLUS_COL
from .selectable import Lateral as Lateral
from .selectable import Select as Select
from .selectable import Selectable as Selectable
from .selectable import Subquery as Subquery
from .selectable import subquery as subquery
from .selectable import TableClause as TableClause
from .selectable import TableSample as TableSample
from .selectable import TableValuedAlias as TableValuedAlias
from .selectable import Values as Values
from .traversals import CacheKey as CacheKey

all_ = CollectionAggregate._create_all
any_ = CollectionAggregate._create_any
and_ = BooleanClauseList.and_
alias = Alias._factory
tablesample = TableSample._factory
lateral = Lateral._factory
or_ = BooleanClauseList.or_
bindparam = BindParameter
select = Select._create
text = TextClause._create_text
table = TableClause
column = ColumnClause
over = Over
within_group = WithinGroup
label = Label
case = Case
cast = Cast
cte = CTE._factory
values = Values
extract = Extract
tuple_ = Tuple
except_ = CompoundSelect._create_except
except_all = CompoundSelect._create_except_all
intersect = CompoundSelect._create_intersect
intersect_all = CompoundSelect._create_intersect_all
union = CompoundSelect._create_union
union_all = CompoundSelect._create_union_all
exists = Exists
nulls_first = UnaryExpression._create_nulls_first
nullsfirst = UnaryExpression._create_nulls_first
nulls_last = UnaryExpression._create_nulls_last
nullslast = UnaryExpression._create_nulls_last
asc = UnaryExpression._create_asc
desc = UnaryExpression._create_desc
distinct = UnaryExpression._create_distinct
type_coerce = TypeCoerce
true = True_._instance
false = False_._instance
null = Null._instance
join = Join._create_join
outerjoin = Join._create_outerjoin
insert = Insert
update = Update
delete = Delete
funcfilter = FunctionFilter



from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import overload
from typing import Set
from typing import Type
from typing import TypeVar
from typing import Union

from typing_extensions import Literal

from . import functions
from . import roles
from . import sqltypes
from . import type_api
from . import visitors
from .base import DedupeColumnCollection
from .base import DialectKWArgs
from .base import SchemaEventTarget
from .elements import ClauseElement
from .elements import ColumnClause
from .elements import TextClause
from .selectable import TableClause
from .. import util
from ..engine import Connection
from ..engine import Engine
from ..util import langhelpers

RETAIN_SCHEMA: langhelpers._symbol
BLANK_SCHEMA: langhelpers._symbol
NULL_UNSPECIFIED: langhelpers._symbol

_T = TypeVar("_T")

_TE = TypeVar("_TE", bound=type_api.TypeEngine[Any])
_TAB = TypeVar("_TAB", bound=Table)
_CO = TypeVar("_CO", bound=Column[Any])
_FK = TypeVar("_FK", bound=ForeignKey)
_FKC = TypeVar("_FKC", bound=ForeignKeyConstraint)
_CT = TypeVar("_CT", bound=Constraint)
_CCC = TypeVar("_CCC", bound=ColumnCollectionConstraint)
_CC = TypeVar("_CC", bound=CheckConstraint)
_IDX = TypeVar("_IDX", bound=Index)
_CP = TypeVar("_CP", bound=Computed)
_ID = TypeVar("_ID", bound=Identity)

class SchemaItem(SchemaEventTarget, visitors.Visitable):
    __visit_name__: str = ...
    create_drop_stringify_dialect: str = ...
    @util.memoized_property
    def info(self) -> Dict[Any, Any]: ...

class Table(DialectKWArgs, SchemaItem, TableClause):
    __visit_name__: str = ...
    metadata: MetaData = ...
    schema: Optional[str] = ...
    indexes: Set[Index] = ...
    constraints: Set[Constraint] = ...
    foreign_keys: Set[ForeignKey] = ...
    primary_key: PrimaryKeyConstraint = ...
    fullname: str = ...
    implicit_returning: bool = ...
    comment: Optional[Any] = ...
    def __new__(cls: Type[_TAB], *args: Any, **kw: Any) -> _TAB: ...
    def __init__(self, *args: Any, **kw: Any) -> None: ...
    @property
    def foreign_key_constraints(self) -> Set[ForeignKeyConstraint]: ...
    @property
    def key(self) -> util.text_type: ...
    @property
    def bind(self) -> Optional[Union[Engine, Connection]]: ...
    def add_is_dependent_on(self, table: Any) -> None: ...
    def append_column(  # type: ignore[override]
        self, column: Any, replace_existing: bool = ...
    ) -> None: ...
    def append_constraint(self, constraint: Any) -> None: ...
    def exists(
        self, bind: Optional[Union[Engine, Connection]] = ...
    ) -> bool: ...
    def create(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        checkfirst: bool = ...,
    ) -> None: ...
    def drop(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        checkfirst: bool = ...,
    ) -> None: ...
    def tometadata(
        self,
        metadata: MetaData,
        schema: Optional[Union[langhelpers._symbol, util.text_type]] = ...,
        referred_schema_fn: Optional[
            Callable[
                [Table, util.text_type, ForeignKeyConstraint, util.text_type],
                util.text_type,
            ]
        ] = ...,
        name: Optional[util.text_type] = ...,
    ) -> Table: ...
    def to_metadata(
        self,
        metadata: MetaData,
        schema: Optional[Union[langhelpers._symbol, util.text_type]] = ...,
        referred_schema_fn: Optional[
            Callable[
                [Table, util.text_type, ForeignKeyConstraint, util.text_type],
                util.text_type,
            ]
        ] = ...,
        name: Optional[util.text_type] = ...,
    ) -> Table: ...

class Column(DialectKWArgs, SchemaItem, ColumnClause[_TE]):
    __visit_name__: str = ...
    inherit_cache: bool = ...
    key: Optional[str] = ...
    primary_key: bool = ...
    nullable: bool = ...
    default: Optional[Any] = ...
    server_default: Optional[Union[FetchedValue, str, TextClause]] = ...
    server_onupdate: Optional[FetchedValue] = ...
    index: Optional[bool] = ...
    unique: Optional[bool] = ...
    system: bool = ...
    doc: Optional[str] = ...
    onupdate: Optional[Any] = ...
    autoincrement: Union[bool, Literal["auto", "ignore_fk"]] = ...
    constraints: Set[Constraint] = ...
    foreign_keys: Set[ForeignKey] = ...
    comment: Optional[str] = ...
    computed: Optional[Computed] = ...
    identity: Optional[Identity] = ...
    @overload
    def __init__(
        self: Column[sqltypes.NullType],
        __name: str,
        *args: SchemaEventTarget,
        autoincrement: Union[bool, Literal["auto", "ignore_fk"]] = ...,
        default: Optional[Any] = ...,
        doc: Optional[str] = ...,
        key: Optional[str] = ...,
        index: Optional[bool] = ...,
        info: Mapping[Any, Any] = ...,
        nullable: bool = ...,
        onupdate: Optional[Any] = ...,
        primary_key: bool = ...,
        server_default: Optional[Union[FetchedValue, str, TextClause]] = ...,
        server_onupdate: Optional[FetchedValue] = ...,
        quote: Optional[bool] = ...,
        unique: Optional[bool] = ...,
        system: bool = ...,
        comment: Optional[str] = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self: Column[sqltypes.NullType],
        *args: SchemaEventTarget,
        autoincrement: Union[bool, Literal["auto", "ignore_fk"]] = ...,
        default: Optional[Any] = ...,
        doc: Optional[str] = ...,
        key: Optional[str] = ...,
        index: Optional[bool] = ...,
        info: Mapping[Any, Any] = ...,
        nullable: bool = ...,
        onupdate: Optional[Any] = ...,
        primary_key: bool = ...,
        server_default: Optional[Union[FetchedValue, str, TextClause]] = ...,
        server_onupdate: Optional[FetchedValue] = ...,
        quote: Optional[bool] = ...,
        unique: Optional[bool] = ...,
        system: bool = ...,
        comment: Optional[str] = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self,
        __name: str,
        __type: Union[_TE, Type[_TE]],
        *args: SchemaEventTarget,
        autoincrement: Union[bool, Literal["auto", "ignore_fk"]] = ...,
        default: Optional[Any] = ...,
        doc: Optional[str] = ...,
        key: Optional[str] = ...,
        index: Optional[bool] = ...,
        info: Mapping[Any, Any] = ...,
        nullable: bool = ...,
        onupdate: Optional[Any] = ...,
        primary_key: bool = ...,
        server_default: Optional[Union[FetchedValue, str, TextClause]] = ...,
        server_onupdate: Optional[FetchedValue] = ...,
        quote: Optional[bool] = ...,
        unique: Optional[bool] = ...,
        system: bool = ...,
        comment: Optional[str] = ...,
        **kwargs: Any,
    ) -> None: ...
    @overload
    def __init__(
        self,
        __type: Union[_TE, Type[_TE]],
        *args: SchemaEventTarget,
        autoincrement: Union[bool, Literal["auto", "ignore_fk"]] = ...,
        default: Optional[Any] = ...,
        doc: Optional[str] = ...,
        key: Optional[str] = ...,
        index: Optional[bool] = ...,
        info: Mapping[Any, Any] = ...,
        nullable: bool = ...,
        onupdate: Optional[Any] = ...,
        primary_key: bool = ...,
        server_default: Optional[Union[FetchedValue, str, TextClause]] = ...,
        server_onupdate: Optional[FetchedValue] = ...,
        quote: Optional[bool] = ...,
        unique: Optional[bool] = ...,
        system: bool = ...,
        comment: Optional[str] = ...,
        **kwargs: Any,
    ) -> None: ...
    def references(self, column: Column[Any]) -> bool: ...
    def append_foreign_key(self, fk: ForeignKey) -> None: ...
    def copy(self: _CO, **kw: Any) -> _CO: ...

class ForeignKey(DialectKWArgs, SchemaItem):
    __visit_name__: str = ...
    constraint: Optional[ForeignKeyConstraint] = ...
    parent: Optional[Column[Any]] = ...
    use_alter: bool = ...
    name: Optional[str] = ...
    onupdate: Optional[str] = ...
    ondelete: Optional[str] = ...
    deferrable: Optional[bool] = ...
    initially: Optional[str] = ...
    link_to_name: bool = ...
    match: Optional[str] = ...
    def __init__(
        self,
        column: Union[Column[Any], str],
        _constraint: Optional[ForeignKeyConstraint] = ...,
        use_alter: bool = ...,
        name: Optional[str] = ...,
        onupdate: Optional[str] = ...,
        ondelete: Optional[str] = ...,
        deferrable: Optional[bool] = ...,
        initially: Optional[str] = ...,
        link_to_name: bool = ...,
        match: Optional[str] = ...,
        info: Optional[Dict[Any, Any]] = ...,
        **dialect_kw: Any,
    ) -> None: ...
    def copy(self: _FK, schema: Optional[str] = ...) -> _FK: ...
    @property
    def target_fullname(self) -> str: ...
    def references(self, table: Table) -> bool: ...
    def get_referent(self, table: Table) -> Column[Any]: ...
    @util.memoized_property
    def column(self) -> Column[Any]: ...

class DefaultGenerator(SchemaItem):
    __visit_name__: str = ...
    is_sequence: bool = ...
    is_server_default: bool = ...
    column: Optional[Column[Any]] = ...
    for_update: bool = ...
    def __init__(self, for_update: bool = ...) -> None: ...
    def execute(
        self, bind: Optional[Union[Engine, Connection]] = ...
    ) -> Any: ...
    @property
    def bind(self) -> Optional[Union[Engine, Connection]]: ...

class ColumnDefault(DefaultGenerator):
    arg: Any = ...
    def __init__(self, arg: Any, **kwargs: Any) -> None: ...
    @util.memoized_property
    def is_callable(self) -> bool: ...
    @util.memoized_property
    def is_clause_element(self) -> bool: ...
    @util.memoized_property
    def is_scalar(self) -> bool: ...

class IdentityOptions:
    start: Optional[int] = ...
    increment: Optional[int] = ...
    minvalue: Optional[int] = ...
    maxvalue: Optional[int] = ...
    nominvalue: Optional[bool] = ...
    nomaxvalue: Optional[bool] = ...
    cycle: Optional[bool] = ...
    cache: Optional[int] = ...
    order: Optional[bool] = ...
    def __init__(
        self,
        start: Optional[int] = ...,
        increment: Optional[int] = ...,
        minvalue: Optional[int] = ...,
        maxvalue: Optional[int] = ...,
        nominvalue: Optional[bool] = ...,
        nomaxvalue: Optional[bool] = ...,
        cycle: Optional[bool] = ...,
        cache: Optional[int] = ...,
        order: Optional[bool] = ...,
    ) -> None: ...

class Sequence(
    IdentityOptions,
    roles.StatementRole,
    DefaultGenerator,
    Generic[_TE],
):
    __visit_name__: str = ...
    is_sequence: bool = ...
    name: str = ...
    optional: bool = ...
    schema: Optional[Union[str, langhelpers._symbol]] = ...
    metadata: Optional[MetaData] = ...
    data_type: Optional[_TE] = ...
    @overload
    def __init__(
        self: Sequence[sqltypes.Integer],
        name: str,
        start: Optional[int] = ...,
        increment: Optional[int] = ...,
        minvalue: Optional[int] = ...,
        maxvalue: Optional[int] = ...,
        nominvalue: Optional[bool] = ...,
        nomaxvalue: Optional[bool] = ...,
        cycle: Optional[bool] = ...,
        schema: Optional[str] = ...,
        cache: Optional[int] = ...,
        order: Optional[bool] = ...,
        data_type: None = ...,
        optional: bool = ...,
        quote: Optional[bool] = ...,
        metadata: Optional[MetaData] = ...,
        quote_schema: Optional[bool] = ...,
        for_update: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        name: str,
        start: int,
        increment: Optional[int],
        minvalue: Optional[int],
        maxvalue: Optional[int],
        nominvalue: Optional[bool],
        nomaxvalue: Optional[bool],
        cycle: Optional[bool],
        schema: Optional[str],
        cache: Optional[int],
        order: Optional[bool],
        data_type: Union[_TE, Type[_TE]],
        optional: bool = ...,
        quote: Optional[bool] = ...,
        metadata: Optional[MetaData] = ...,
        quote_schema: Optional[bool] = ...,
        for_update: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        name: str,
        *,
        start: Optional[int] = ...,
        increment: Optional[int] = ...,
        minvalue: Optional[int] = ...,
        maxvalue: Optional[int] = ...,
        nominvalue: Optional[bool] = ...,
        nomaxvalue: Optional[bool] = ...,
        cycle: Optional[bool] = ...,
        schema: Optional[str] = ...,
        cache: Optional[int] = ...,
        order: Optional[bool] = ...,
        data_type: Union[_TE, Type[_TE]],
        optional: bool = ...,
        quote: Optional[bool] = ...,
        metadata: Optional[MetaData] = ...,
        quote_schema: Optional[bool] = ...,
        for_update: bool = ...,
    ) -> None: ...
    @util.memoized_property
    def is_callable(self) -> bool: ...
    @util.memoized_property
    def is_clause_element(self) -> bool: ...
    def next_value(self) -> functions.next_value[_TE]: ...
    @property
    def bind(self) -> Optional[Union[Engine, Connection]]: ...
    def create(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        checkfirst: bool = ...,
    ) -> None: ...
    def drop(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        checkfirst: bool = ...,
    ) -> None: ...

class FetchedValue(SchemaEventTarget):
    is_server_default: bool = ...
    reflected: bool = ...
    has_argument: bool = ...
    is_clause_element: bool = ...
    for_update: bool = ...
    def __init__(self, for_update: bool = ...) -> None: ...

class DefaultClause(FetchedValue):
    has_argument: bool = ...
    arg: Any = ...
    reflected: bool = ...
    def __init__(
        self, arg: Any, for_update: bool = ..., _reflected: bool = ...
    ) -> None: ...

class Constraint(DialectKWArgs, SchemaItem):
    __visit_name__: str = ...
    name: Optional[str] = ...
    deferrable: Optional[bool] = ...
    initially: Optional[str] = ...
    def __init__(
        self,
        name: Optional[str] = ...,
        deferrable: Optional[bool] = ...,
        initially: Optional[str] = ...,
        _create_rule: Optional[Any] = ...,
        info: Optional[Dict[Any, Any]] = ...,
        _type_bound: bool = ...,
        **dialect_kw: Any,
    ) -> None: ...
    @property
    def table(self) -> Table: ...
    def copy(self: _CT, **kw: Any) -> _CT: ...

class ColumnCollectionMixin:
    columns: DedupeColumnCollection[Column[Any]] = ...
    def __init__(
        self, *columns: Union[str, Column[Any]], **kw: Any
    ) -> None: ...

class ColumnCollectionConstraint(ColumnCollectionMixin, Constraint):
    def __init__(
        self, *columns: Union[str, Column[Any]], **kw: Any
    ) -> None: ...
    def __contains__(self, x: Any) -> bool: ...
    def copy(
        self: _CCC, target_table: Optional[Table] = ..., **kw: Any
    ) -> _CCC: ...
    def contains_column(self, col: Column[Any]) -> bool: ...
    def __iter__(self) -> Iterator[Column[Any]]: ...
    def __len__(self) -> int: ...

class CheckConstraint(ColumnCollectionConstraint):
    __visit_name__: str = ...
    sqltext: ClauseElement = ...
    def __init__(
        self,
        sqltext: Union[str, ClauseElement],
        name: Optional[str] = ...,
        deferrable: Optional[bool] = ...,
        initially: Optional[str] = ...,
        table: Optional[Table] = ...,
        info: Optional[Dict[Any, Any]] = ...,
        _create_rule: Optional[Any] = ...,
        _autoattach: bool = ...,
        _type_bound: bool = ...,
        **kw: Any,
    ) -> None: ...
    @property
    def is_column_level(self) -> bool: ...
    def copy(
        self: _CC, target_table: Optional[Table] = ..., **kw: Any
    ) -> _CC: ...

class ForeignKeyConstraint(ColumnCollectionConstraint):
    __visit_name__: str = ...
    onupdate: Optional[str] = ...
    ondelete: Optional[str] = ...
    link_to_name: bool = ...
    use_alter: bool = ...
    match: Optional[str] = ...
    elements: List[ForeignKey] = ...
    def __init__(
        self,
        columns: Iterable[str],
        refcolumns: Iterable[Union[str, Column[Any]]],
        name: Optional[str] = ...,
        onupdate: Optional[str] = ...,
        ondelete: Optional[str] = ...,
        deferrable: Optional[bool] = ...,
        initially: Optional[str] = ...,
        use_alter: bool = ...,
        link_to_name: bool = ...,
        match: Optional[str] = ...,
        table: Optional[Table] = ...,
        info: Optional[Dict[Any, Any]] = ...,
        **dialect_kw: Any,
    ) -> None: ...
    @property
    def referred_table(self) -> Table: ...
    @property
    def column_keys(self) -> List[str]: ...
    def copy(  # type: ignore[override]
        self: _FKC,
        schema: Optional[str] = ...,
        target_table: Optional[Table] = ...,
        **kw: Any,
    ) -> _FKC: ...

class PrimaryKeyConstraint(ColumnCollectionConstraint):
    __visit_name__: str = ...
    def __init__(
        self, *columns: Union[str, Column[Any]], **kw: Any
    ) -> None: ...
    @property
    def columns_autoinc_first(self) -> List[Column[Any]]: ...

class UniqueConstraint(ColumnCollectionConstraint):
    __visit_name__: str = ...

class Index(DialectKWArgs, ColumnCollectionMixin, SchemaItem):
    __visit_name__: str = ...
    table: Optional[Table] = ...
    name: str = ...
    unique: bool = ...
    expressions: List[Column[Any]] = ...
    def __init__(
        self, name: str, *expressions: ColumnClause[Any], **kw: Any
    ) -> None: ...
    @property
    def bind(self) -> Optional[Union[Engine, Connection]]: ...
    def create(
        self: _IDX,
        bind: Optional[Union[Engine, Connection]] = ...,
        checkfirst: bool = ...,
    ) -> _IDX: ...
    def drop(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        checkfirst: bool = ...,
    ) -> None: ...

DEFAULT_NAMING_CONVENTION: util.immutabledict[str, str]

class MetaData(SchemaItem):
    __visit_name__: str = ...
    tables: util.FacadeDict[str, Table] = ...
    schema: Optional[str] = ...
    naming_convention: Dict[Any, Any] = ...
    bind: Optional[Union[Engine, Connection]] = ...
    def __init__(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        schema: Optional[str] = ...,
        quote_schema: Optional[bool] = ...,
        naming_convention: Optional[Dict[str, str]] = ...,
        info: Optional[Dict[Any, Any]] = ...,
    ) -> None: ...
    def __contains__(self, table_or_key: Any) -> bool: ...
    def is_bound(self) -> bool: ...
    def clear(self) -> None: ...
    def remove(self, table: Table) -> None: ...
    @property
    def sorted_tables(self) -> List[Table]: ...
    def reflect(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        schema: Optional[Any] = ...,
        views: bool = ...,
        only: Optional[Any] = ...,
        extend_existing: bool = ...,
        autoload_replace: bool = ...,
        resolve_fks: bool = ...,
        **dialect_kwargs: Any,
    ) -> None: ...
    def create_all(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        tables: Optional[List[Table]] = ...,
        checkfirst: bool = ...,
    ) -> None: ...
    def drop_all(
        self,
        bind: Optional[Union[Engine, Connection]] = ...,
        tables: Optional[List[Table]] = ...,
        checkfirst: bool = ...,
    ) -> None: ...

class ThreadLocalMetaData(MetaData):
    __visit_name__: str = ...
    context: util.threading.local = ...
    bind: Optional[Union[Engine, Connection]] = ...
    def __init__(self) -> None: ...
    def is_bound(self) -> bool: ...
    def dispose(self) -> None: ...

class Computed(FetchedValue, SchemaItem):
    __visit_name__: str = ...
    sqltext: ClauseElement = ...
    persisted: Optional[bool] = ...
    column: Optional[Column[Any]] = ...
    def __init__(
        self,
        sqltext: Union[str, ClauseElement],
        persisted: Optional[bool] = ...,
    ) -> None: ...
    def copy(
        self: _CP, target_table: Optional[Table] = ..., **kw: Any
    ) -> _CP: ...

class Identity(IdentityOptions, FetchedValue, SchemaItem):
    __visit_name__: str = ...
    always: bool = ...
    on_null: Optional[bool] = ...
    column: Optional[Column[Any]] = ...
    def __init__(
        self,
        always: bool = ...,
        on_null: Optional[bool] = ...,
        start: Optional[int] = ...,
        increment: Optional[int] = ...,
        minvalue: Optional[int] = ...,
        maxvalue: Optional[int] = ...,
        nominvalue: Optional[bool] = ...,
        nomaxvalue: Optional[bool] = ...,
        cycle: Optional[bool] = ...,
        cache: Optional[int] = ...,
        order: Optional[bool] = ...,
    ) -> None: ...
    def copy(self: _ID, **kw: Any) -> _ID: ...



# fmt: off
from typing import Any
from typing import Optional

from . import attributes as attributes
from . import path_registry as path_registry
from . import strategy_options as strategy_options
from .util import state_str as state_str
from .. import future as future
from .. import util as util
from ..engine import result_tuple as result_tuple
from ..engine.result import ChunkedIteratorResult as ChunkedIteratorResult
from ..engine.result import FrozenResult as FrozenResult
from ..engine.result import SimpleResultMetaData as SimpleResultMetaData
from ..sql.selectable import LABEL_STYLE_TABLENAME_PLUS_COL as LABEL_STYLE_TABLENAME_PLUS_COL
# fmt: on

def instances(cursor: Any, context: Any): ...
def merge_frozen_result(
    session: Any, statement: Any, frozen_result: Any, load: bool = ...
): ...
def merge_result(query: Any, iterator: Any, load: bool = ...): ...
def get_from_identity(session: Any, mapper: Any, key: Any, passive: Any): ...
def load_on_ident(
    session: Any,
    statement: Any,
    key: Any,
    load_options: Optional[Any] = ...,
    refresh_state: Optional[Any] = ...,
    with_for_update: Optional[Any] = ...,
    only_load_props: Optional[Any] = ...,
    no_autoflush: bool = ...,
    bind_arguments: Any = ...,
    execution_options: Any = ...,
): ...
def load_on_pk_identity(
    session: Any,
    statement: Any,
    primary_key_identity: Any,
    load_options: Optional[Any] = ...,
    refresh_state: Optional[Any] = ...,
    with_for_update: Optional[Any] = ...,
    only_load_props: Optional[Any] = ...,
    identity_token: Optional[Any] = ...,
    no_autoflush: bool = ...,
    bind_arguments: Any = ...,
    execution_options: Any = ...,
): ...

class PostLoad:
    loaders: Any = ...
    states: Any = ...
    load_keys: Any = ...
    def __init__(self) -> None: ...
    def add_state(self, state: Any, overwrite: Any) -> None: ...
    def invoke(self, context: Any, path: Any) -> None: ...
    @classmethod
    def for_context(cls, context: Any, path: Any, only_load_props: Any): ...
    @classmethod
    def path_exists(self, context: Any, path: Any, key: Any): ...
    @classmethod
    def callable_for_path(
        cls,
        context: Any,
        path: Any,
        limit_to_mapper: Any,
        token: Any,
        loader_callable: Any,
        *arg: Any,
        **kw: Any,
    ) -> None: ...

def load_scalar_attributes(
    mapper: Any, state: Any, attribute_names: Any, passive: Any
) -> None: ...



from typing import Any
from typing import Optional

from . import base as base
from . import collections as collections
from . import exc as exc
from . import interfaces as interfaces
from . import state as state
from .. import util as util
from ..util import HasMemoized as HasMemoized

DEL_ATTR: Any

class ClassManager(HasMemoized, dict):
    MANAGER_ATTR: Any = ...
    STATE_ATTR: Any = ...
    expired_attribute_loader: Any = ...
    init_method: Any = ...
    factory: Any = ...
    mapper: Any = ...
    declarative_scan: Any = ...
    registry: Any = ...
    @property
    def deferred_scalar_loader(self): ...
    @deferred_scalar_loader.setter
    def deferred_scalar_loader(self, obj: Any) -> None: ...
    class_: Any = ...
    info: Any = ...
    new_init: Any = ...
    local_attrs: Any = ...
    originals: Any = ...
    def __init__(self, class_: Any) -> None: ...
    def __hash__(self) -> Any: ...
    def __eq__(self, other: Any) -> Any: ...
    @property
    def is_mapped(self): ...
    def mapper(self) -> None: ...
    def manage(self) -> None: ...
    def manager_getter(self): ...
    def state_getter(self): ...
    def dict_getter(self): ...
    def instrument_attribute(
        self, key: Any, inst: Any, propagated: bool = ...
    ) -> None: ...
    def subclass_managers(self, recursive: Any) -> None: ...
    def post_configure_attribute(self, key: Any) -> None: ...
    def uninstrument_attribute(
        self, key: Any, propagated: bool = ...
    ) -> None: ...
    def unregister(self) -> None: ...
    def install_descriptor(self, key: Any, inst: Any) -> None: ...
    def uninstall_descriptor(self, key: Any) -> None: ...
    def install_member(self, key: Any, implementation: Any) -> None: ...
    def uninstall_member(self, key: Any) -> None: ...
    def instrument_collection_class(self, key: Any, collection_class: Any): ...
    def initialize_collection(self, key: Any, state: Any, factory: Any): ...
    def is_instrumented(self, key: Any, search: bool = ...): ...
    def get_impl(self, key: Any): ...
    @property
    def attributes(self): ...
    def new_instance(self, state: Optional[Any] = ...): ...
    def setup_instance(
        self, instance: Any, state: Optional[Any] = ...
    ) -> None: ...
    def teardown_instance(self, instance: Any) -> None: ...
    def has_state(self, instance: Any): ...
    def has_parent(self, state: Any, key: Any, optimistic: bool = ...): ...
    def __bool__(self): ...
    __nonzero__: Any = ...

class _SerializeManager:
    class_: Any = ...
    def __init__(self, state: Any, d: Any) -> None: ...
    def __call__(self, state: Any, inst: Any, state_dict: Any) -> None: ...

class InstrumentationFactory:
    def create_manager_for_cls(self, class_: Any): ...
    def unregister(self, class_: Any) -> None: ...

instance_state: Any

instance_dict: Any
manager_of_class = base.manager_of_class

def register_class(
    class_: Any,
    finalize: bool = ...,
    mapper: Optional[Any] = ...,
    registry: Optional[Any] = ...,
    declarative_scan: Optional[Any] = ...,
    expired_attribute_loader: Optional[Any] = ...,
    init_method: Optional[Any] = ...,
): ...
def unregister_class(class_: Any) -> None: ...
def is_instrumented(instance: Any, key: Any): ...



from typing import Any
from typing import Optional

from .. import exc as exc
from .. import util as util

class _EventKey:
    target: Any = ...
    identifier: Any = ...
    fn: Any = ...
    fn_key: Any = ...
    fn_wrap: Any = ...
    dispatch_target: Any = ...
    def __init__(
        self,
        target: Any,
        identifier: Any,
        fn: Any,
        dispatch_target: Any,
        _fn_wrap: Optional[Any] = ...,
    ) -> None: ...
    def with_wrapper(self, fn_wrap: Any): ...
    def with_dispatch_target(self, dispatch_target: Any): ...
    def listen(self, *args: Any, **kw: Any) -> None: ...
    def remove(self) -> None: ...
    def contains(self): ...
    def base_listen(
        self,
        propagate: bool = ...,
        insert: bool = ...,
        named: bool = ...,
        retval: Optional[Any] = ...,
        asyncio: bool = ...,
    ) -> None: ...
    def append_to_list(self, owner: Any, list_: Any): ...
    def remove_from_list(self, owner: Any, list_: Any) -> None: ...
    def prepend_to_list(self, owner: Any, list_: Any): ...



from typing import Any

from mypy.plugin import Callable as Callable
from mypy.plugin import CallableType as CallableType
from mypy.plugin import ClassDefContext as ClassDefContext
from mypy.plugin import FunctionSigContext as FunctionSigContext
from mypy.plugin import Optional as Optional
from mypy.plugin import Plugin

cls_decorator_names_we_care_about: Any
classes_w_clsdec_methods: Any

class CustomPlugin(Plugin):
    def get_function_signature_hook(
        self, fullname: str
    ) -> Optional[Callable[[FunctionSigContext], CallableType]]: ...
    def get_class_decorator_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]: ...
    def get_customize_class_mro_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]: ...

def fill_in_decorators(ctx: ClassDefContext) -> None: ...
def cls_decorator_hook(ctx: ClassDefContext) -> None: ...
def plugin(version: str) -> Any: ...



from typing import Any
from typing import Dict
from typing import Type

from ... import MetaData

def instrument_declarative(
    cls: Type[Any], cls_registry: Dict[str, Type[Any]], metadata: MetaData
) -> Any: ...

class ConcreteBase:
    @classmethod
    def __declare_first__(cls) -> None: ...

class AbstractConcreteBase(ConcreteBase):
    __no_table__: bool = ...
    @classmethod
    def __declare_first__(cls) -> None: ...

class DeferredReflection:
    @classmethod
    def prepare(cls, engine: Any) -> None: ...
